import ctypes
import logging
import struct

import cbor2

from . import defines
from .messages.message import Message
from .messages.option import Option
from .messages.request import Request
from .messages.response import Response
from .serializer import Serializer

__author__ = 'Giacomo Tanganelli'

logger = logging.getLogger(__name__)


def _extract_message_size(data: bytes):
    """Read out the full length of a CoAP messsage represented by data.

    Returns None if data is too short to read the (full) length.

    The number returned is the number of bytes that has to be read into data to
    start reading the next message; it consists of a constant term, the token
    length and the extended length of options-plus-payload."""

    if not data:
        return None

    l = data[0] >> 4
    tokenoffset = 2
    tkl = data[0] & 0x0f

    if l >= 13:
        if l == 13:
            extlen = 1
            offset = 13
        elif l == 14:
            extlen = 2
            offset = 269
        else:
            extlen = 4
            offset = 65805
        if len(data) < extlen + 1:
            return None
        tokenoffset = 2 + extlen
        l = int.from_bytes(data[1:1 + extlen], "big") + offset
    return tokenoffset, tkl, l


def _decode_message(data: bytes) -> Message:
    tokenoffset, tkl, _ = _extract_message_size(data)
    if tkl > 8:
        raise error.UnparsableMessage("Overly long token")
    code = data[tokenoffset - 1]
    token = data[tokenoffset:tokenoffset + tkl]

    msg = Message(code=code, token=token)

    msg.payload = msg.opt.decode(data[tokenoffset + tkl:])

    return msg


def _encode_length(l: int):
    if l < 13:
        return (l, b"")
    elif l < 269:
        return (13, (l - 13).to_bytes(1, 'big'))
    elif l < 65805:
        return (14, (l - 269).to_bytes(2, 'big'))
    else:
        return (15, (l - 65805).to_bytes(4, 'big'))


def _serialize(msg: Message) -> bytes:
    data = [msg.opt.encode()]
    if msg.payload:
        data += [b'\xff', msg.payload]
    data = b"".join(data)
    l, extlen = _encode_length(len(data))

    tkl = len(msg.token)
    if tkl > 8:
        raise ValueError("Overly long token")

    return b"".join((
        bytes(((l << 4) | tkl,)),
        extlen,
        bytes((msg.code,)),
        msg.token,
        data
    ))


class SerializerTcp(Serializer):
    """
    Serializer class to serialize and deserialize CoAP message to/from udp streams.
    """

    @staticmethod
    def deserialize(datagram, source):
        """
        De-serialize a stream of byte to a message.

        :param datagram: the incoming udp message
        :param source: the source address and port (ip, port)
        :return: the message
        :rtype: Message
        """
        try:
            tokenoffset, tkl, _ = _extract_message_size(datagram)
            if tkl > 8:
                raise UnicodeDecodeError("Overly long token")
            code = datagram[tokenoffset - 1]
            token = datagram[tokenoffset:tokenoffset + tkl]

            # fmt = "!BBH"
            # pos = struct.calcsize(fmt)
            # s = struct.Struct(fmt)
            # values = s.unpack_from(datagram)
            # first = values[0]
            # code = values[1]
            # mid = values[2]
            # version = (first & 0xC0) >> 6
            # message_type = (first & 0x30) >> 4
            # token_length = (first & 0x0F)
            if Serializer.is_response(code):
                message = Response()
                message.code = code
            elif Serializer.is_request(code):
                message = Request()
                message.code = code
            else:
                message = Message()
                message.code = code
            message.source = source
            message.destination = None
            # message.version = version
            # message.type = message_type
            # message.mid = mid
            # if token_length > 0:
            #     message.token = datagram[pos:pos + token_length]
            # else:
            #     message.token = None
            message.token = token

            current_option = 0
            values = datagram[tokenoffset + tkl:]
            length_packet = len(values)
            pos = 0
            while pos < length_packet:
                next_byte = struct.unpack("B", values[pos].to_bytes(1, "big"))[0]
                pos += 1
                if next_byte != int(defines.PAYLOAD_MARKER):
                    # the first 4 bits of the byte represent the option delta
                    # delta = self._reader.read(4).uint
                    num, option_length, pos = Serializer.read_option_value_len_from_byte(next_byte, pos, values)
                    # logger.debug("option value (delta): %d len: %d", num, option_length)
                    current_option += num
                    # read option
                    try:
                        option_item = defines.OptionRegistry.LIST[current_option]
                    except KeyError:
                        (opt_critical, _, _) = defines.OptionRegistry.get_option_flags(current_option)
                        if opt_critical:
                            raise AttributeError("Critical option %s unknown" % current_option)
                        else:
                            # If the non-critical option is unknown
                            # (vendor-specific, proprietary) - just skip it
                            logger.warning("unrecognized option %d", current_option)
                    else:
                        if option_length == 0:
                            value = None
                        elif option_item.value_type == defines.INTEGER:
                            tmp = values[pos: pos + option_length]
                            value = 0
                            for b in tmp:
                                value = (value << 8) | struct.unpack("B", b.to_bytes(1, "big"))[0]
                        elif option_item.value_type == defines.OPAQUE:
                            tmp = values[pos: pos + option_length]
                            value = tmp
                        else:
                            value = values[pos: pos + option_length]

                        option = Option()
                        option.number = current_option
                        option.value = Serializer.convert_to_raw(current_option, value, option_length)

                        message.add_option(option)
                        if option.number == defines.OptionRegistry.CONTENT_TYPE.number:
                            message.payload_type = option.value
                    finally:
                        pos += option_length
                else:

                    if length_packet <= pos:
                        # log.err("Payload Marker with no payload")
                        raise AttributeError("Packet length %s, pos %s" % (length_packet, pos))
                    message.payload = ""
                    payload = values[pos:]
                    message.payload = payload

                    pos += len(payload)

            return message
        except AttributeError:
            return defines.Codes.BAD_REQUEST.number
        except struct.error:
            return defines.Codes.BAD_REQUEST.number
        except UnicodeDecodeError as e:
            logger.debug(e)
            return defines.Codes.BAD_REQUEST.number

    @staticmethod
    def serialize(message):
        """
        Serialize a message to a udp packet

        :type message: Message
        :param message: the message to be serialized
        :rtype: stream of byte
        :return: the message serialized
        """
        fmt = "!BBH"
        fmt = ""
        # if message.token is None:
        #     tkl = 0
        # else:
        #     tkl = len(message.token)
        # tmp = (defines.VERSION << 2)
        # tmp |= message.type
        # tmp <<= 4
        # tmp |= tkl
        #
        # values = [tmp, message.code, message.mid]
        values = []

        # if message.token is not None and tkl > 0:
        #     fmt += "%ss" % tkl
        #     values.append(message.token)

        options = Serializer.as_sorted_list(message.options)  # already sorted
        lastoptionnumber = 0
        for option in options:

            # write 4-bit option delta
            optiondelta = option.number - lastoptionnumber
            optiondeltanibble = Serializer.get_option_nibble(optiondelta)
            tmp = (optiondeltanibble << defines.OPTION_DELTA_BITS)

            # write 4-bit option length
            optionlength = option.length
            optionlengthnibble = Serializer.get_option_nibble(optionlength)
            tmp |= optionlengthnibble
            fmt += "B"
            values.append(tmp)

            # write extended option delta field (0 - 2 bytes)
            if optiondeltanibble == 13:
                fmt += "B"
                values.append(optiondelta - 13)
            elif optiondeltanibble == 14:
                fmt += "H"
                values.append(optiondelta - 269)

            # write extended option length field (0 - 2 bytes)
            if optionlengthnibble == 13:
                fmt += "B"
                values.append(optionlength - 13)
            elif optionlengthnibble == 14:
                fmt += "H"
                values.append(optionlength - 269)

            # write option value
            if optionlength > 0:
                opt_type = defines.OptionRegistry.LIST[option.number].value_type
                if opt_type == defines.INTEGER:
                    words = Serializer.int_to_words(option.value, optionlength, 8)
                    for num in range(0, optionlength):
                        fmt += "B"
                        values.append(words[num])
                elif opt_type == defines.STRING:
                    fmt += str(len(bytes(option.value, "utf-8"))) + "s"
                    values.append(bytes(option.value, "utf-8"))
                else:  # OPAQUE
                    for b in option.value:
                        fmt += "B"
                        values.append(b)

            # update last option number
            lastoptionnumber = option.number

        payload = message.payload

        if payload is not None and len(payload) > 0:
            # if payload is present and of non-zero length, it is prefixed by
            # an one-byte Payload Marker (0xFF) which indicates the end of
            # options and the start of the payload

            fmt += "B"
            values.append(defines.PAYLOAD_MARKER)

            if isinstance(payload, bytes):
                fmt += str(len(payload)) + "s"
                values.append(payload)

            else:
                # raise ValueError('Not bytes payload')
                # try:
                #     _encoder = encoder[message.content_type]
                # except KeyError:
                #     _encoder =
                _data = string_encode(payload)
                fmt += str(len(_data)) + "s"
                values.append(_data)

        datagram = None
        # if values[1] is None:
        #     values[1] = 0
        # if values[2] is None:
        #     values[2] = 0
        try:
            s = struct.Struct(fmt)
            datagram = ctypes.create_string_buffer(s.size)
            s.pack_into(datagram, 0, *values)
        except struct.error:
            # The .exception method will report on the exception encountered
            # and provide a traceback.
            logger.debug(fmt)
            logger.debug(values)
            logging.exception('Failed to pack structure')

        l, extlen = _encode_length(len(datagram))
        token = message.token if message.token else b''
        tkl = len(token)
        if tkl > 8:
            raise ValueError("Overly long token")
        datagram = b"".join((
            bytes(((l << 4) | tkl,)),
            extlen,
            bytes((message.code,)),
            token,
            datagram
            ))
        return datagram

    @staticmethod
    def is_request(code):
        """
        Checks if is request.

        :return: True, if is request
        """
        return defines.REQUEST_CODE_LOWER_BOUND <= code <= defines.REQUEST_CODE_UPPER_BOUND

    @staticmethod
    def is_response(code):
        """
        Checks if is response.

        :return: True, if is response
        """
        return defines.RESPONSE_CODE_LOWER_BOUND <= code <= defines.RESPONSE_CODE_UPPER_BOUND

    @staticmethod
    def read_option_value_len_from_byte(byte, pos, values):
        """
        Calculates the value and length used in the extended option fields.

        :param byte: 1-byte option header value.
        :return: the value and length, calculated from the header including the extended fields.
        """
        h_nibble = (byte & 0xF0) >> 4
        l_nibble = byte & 0x0F
        value = 0
        length = 0
        if h_nibble <= 12:
            value = h_nibble
        elif h_nibble == 13:
            value = struct.unpack("!B", values[pos].to_bytes(1, "big"))[0] + 13
            pos += 1
        elif h_nibble == 14:
            s = struct.Struct("!H")
            value = s.unpack_from(values[pos:pos + 2])[0] + 269
            pos += 2
        else:
            raise AttributeError("Unsupported option number nibble " + str(h_nibble))

        if l_nibble <= 12:
            length = l_nibble
        elif l_nibble == 13:
            length = struct.unpack("!B", values[pos].to_bytes(1, "big"))[0] + 13
            pos += 1
        elif l_nibble == 14:
            s = struct.Struct("!H")
            length = s.unpack_from(values[pos:pos + 2])[0] + 269
            pos += 2
        else:
            raise AttributeError("Unsupported option length nibble " + str(l_nibble))
        return value, length, pos

    @staticmethod
    def convert_to_raw(number, value, length):
        """
        Get the value of an option as bytes.

        :param number: the option number
        :param value: the option value
        :param length: the option length
        :return: the value of an option as a BitArray
        """

        opt_type = defines.OptionRegistry.LIST[number].value_type

        if length == 0 and opt_type != defines.INTEGER:
            return bytes()
        elif length == 0 and opt_type == defines.INTEGER:
            return 0
        elif opt_type == defines.STRING:
            if isinstance(value, bytes):
                return value.decode("utf-8")
        elif opt_type == defines.OPAQUE:
            if isinstance(value, bytes):
                return value
            else:
                return bytes(value, "utf-8")
        if isinstance(value, tuple):
            value = value[0]
        if isinstance(value, str):
            value = str(value)
        if isinstance(value, str):
            return bytes(value, "utf-8")
        elif isinstance(value, int):
            return value
        else:
            return bytes(value)

    @staticmethod
    def as_sorted_list(options):
        """
        Returns all options in a list sorted according to their option numbers.

        :return: the sorted list
        """
        if len(options) > 0:
            options = sorted(options, key=lambda o: o.number)
        return options

    @staticmethod
    def get_option_nibble(optionvalue):
        """
        Returns the 4-bit option header value.

        :param optionvalue: the option value (delta or length) to be encoded.
        :return: the 4-bit option header value.
         """
        if optionvalue <= 12:
            return optionvalue
        elif optionvalue <= 255 + 13:
            return 13
        elif optionvalue <= 65535 + 269:
            return 14
        else:
            raise AttributeError("Unsupported option delta " + optionvalue)

    @staticmethod
    def int_to_words(int_val, num_words=4, word_size=32):
        """
        Convert a int value to bytes.

        :param int_val: an arbitrary length Python integer to be split up.
            Network byte order is assumed. Raises an IndexError if width of
            integer (in bits) exceeds word_size * num_words.

        :param num_words: number of words expected in return value tuple.

        :param word_size: size/width of individual words (in bits).

        :return: a list of fixed width words based on provided parameters.
        """
        max_int = 2 ** (word_size * num_words) - 1
        max_word_size = 2 ** word_size - 1

        if not 0 <= int_val <= max_int:
            raise AttributeError('integer %r is out of bounds!' % hex(int_val))

        words = []
        for _ in range(num_words):
            word = int_val & max_word_size
            words.append(int(word))
            int_val >>= word_size
        words.reverse()

        return words

    def diffie_hellman_key_exchange(self):
        from cryptography.hazmat.primitives import hashes
        from cryptography.hazmat.primitives.asymmetric import dh
        from cryptography.hazmat.primitives.kdf.hkdf import HKDF
        # Generate some parameters. These can be reused.
        parameters = dh.generate_parameters(generator=2, key_size=2048)
        # Generate a private key for use in the exchange.
        server_private_key = parameters.generate_private_key()
        # In a real handshake the peer is a remote client. For this
        # example we'll generate another local private key though. Note that in
        # a DH handshake both peers must agree on a common set of parameters.
        peer_private_key = parameters.generate_private_key()
        shared_key = server_private_key.exchange(peer_private_key.public_key())
        # Perform key derivation.
        derived_key = HKDF(
            algorithm=hashes.SHA256(),
            length=32,
            salt=None,
            info=b'handshake data',
        ).derive(shared_key)
        # And now we can demonstrate that the handshake performed in the
        # opposite direction gives the same final value
        same_shared_key = peer_private_key.exchange(
            server_private_key.public_key()
        )
        same_derived_key = HKDF(
            algorithm=hashes.SHA256(),
            length=32,
            salt=None,
            info=b'handshake data',
        ).derive(same_shared_key)
        derived_key == same_derived_key


def cbor_loads(payload):
    return cbor2.loads(payload)


def cbor_dumps(payload):
    return cbor2.dumps(payload) if payload else b''


def string_encode(payload):
    try:
        return bytes(payload, "utf-8")
    except Exception as err:
        raise err


def string_decode(payload):
    return cbor2.loads(payload)


encoder = {
    10000: cbor_dumps
}
decoder = {
    10000: cbor_loads
}



__author__ = 'Mikhail Razgovorov'

import asyncio
import logging
import socket
from ..utils import calc_family_by_address
from .endpoint import Endpoint
from ..coap_tcp_protocol import CoapTcpProtocol
from ..utils import str_append_hash
from ..serializer_tcp import SerializerTcp

logger = logging.getLogger(__name__)


class TcpCoapEndpoint(Endpoint):
    """
    Class to handle the EndPoint.
    """
    scheme = 'coap+tcp'
    serializer = SerializerTcp

    def __init__(self, **kwargs):
        """
        Data structure that represent a EndPoint
        """
        super().__init__(**kwargs)
        self.pool = {}
        self._sock = None
        self._transport = None
        self._server = None
        self._protocol = None
        self._multicast_addresses = None
        self._connected = None
        self.destination = None
        self.csm = asyncio.Future()

    def is_closing(self):
        if self.is_client:
            if not self.transport:
                return True
            return self.transport.is_closing()
        else:
            if not self._server:
                return True
            return self._server.is_serving()

    @property
    def sock(self):
        return self._sock

    @property
    def transport(self):
        return self._transport

    # @transport.setter
    # def transport(self, value):
    #     self._transport = value

    @property
    def protocol(self):
        return self._protocol

    @protocol.setter
    def protocol(self, value):
        self._protocol = value

    async def start_client(self, server, address):
        self._transport, self._protocol = await server.loop.create_connection(
            lambda: CoapTcpProtocol(server, self),
            address[0], address[1])
        self.csm = await asyncio.wait_for(self.csm, 30)
        self.is_client = True
        self._family, self.destination = calc_family_by_address(address)
        new_address = self.transport.get_extra_info('sockname')
        if self._address and self._address != new_address:
            self._address = new_address
        return self

    @classmethod
    def init_by_socket(cls, sock: socket, is_multicast: bool = False):
        """
        Initialize the endpoint over a ready socket.

        :param sock: socket
        :param is_multicast: if socket is a multicast
        """
        self = cls()
        self._is_multicast = is_multicast
        self._sock = sock
        return self

    def send(self, data, address, **kwargs):
        pool_id = str_append_hash(address[0], address[1])
        protocol = self.pool[pool_id]
        protocol.send(data)

    def _init_unicast(self, address):
        self._family, self._address = calc_family_by_address(address)
        if self._family == socket.AF_INET:  # IPv4
            return self.init_unicast_ip4_by_address(address, **self.params)
        elif self._family == socket.AF_INET6:  # IPv6
            return self.init_unicast_ip6_by_address(address, **self.params)

    async def init_unicast(self, server, address):
        try:
            self._init_unicast(address)
        except Exception as err:
            address = (address[0], 0)  # если порт занят сбрасываем
            self._init_unicast(address)

        await self.listen(server)

    def init_unicast_ip4_by_address(self, address, **kwargs):
        self._address = address
        self._family = socket.AF_INET
        self._sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        # self._sock.bind(address)
        # import asyncio
        # loop = asyncio.get_event_loop()
        # loop.create_server()
        return self

    def init_unicast_ip6_by_address(self, address, **kwargs):
        raise NotImplementedError()
        # self._multicast = None
        # self._address = address
        # self._family = socket.AF_INET6
        # self._sock = socket.socket(socket.AF_INET6, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        # self._sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        # self._sock.bind(address)

    async def create_datagram_endpoint(self, server, protocol_factory):
        return await server.loop.create_server(
            lambda: protocol_factory(server, self, is_server=True),
            self.address[0], self.address[1],
            # sock=self._sock
        )

    async def listen(self, server):
        self._server = await self.create_datagram_endpoint(server, CoapTcpProtocol)
        _address = self._server.sockets[0].getsockname()
        source_port = self._address[1]
        if source_port:
            if source_port != _address[1]:
                raise Exception(f'source port {source_port} not installed')
        else:
            self._address = (self._address[0], _address[1])
        logger.debug(f'run {"multicast " if self._multicast else ""}endpoint {_address[0]}:{_address[1]}')
        # _address = socket.getaddrinfo(socket.gethostname(), _address[1], socket.AF_INET, socket.SOCK_DGRAM)[0][4]

    def close(self):
        logger.debug(f'close')
        if self.is_client:
            if self._transport:
                self._transport.close()
            self.pool.pop(str_append_hash(self.address[0], self.address[1]), None)
        else:
            connections = list(self.pool.keys())
            for elem in connections:
                protocol = self.pool.pop(elem, None)
                protocol.close()
            if self._server:
                self._server.close()

    async def restart_transport(self, server):
        self.close()
        await self.init_unicast(server, self.address)




# This file is part of the Python aiocoap library project.
#
# Copyright (c) 2012-2014 Maciej Wasilak <http://sixpinetrees.blogspot.com/>,
#               2013-2014 Christian Amsüss <c.amsuess@energyharvesting.at>
#
# aiocoap is free software, this file is published under the MIT license as
# described in the accompanying LICENSE file.

"""Module in which all meaningful numbers are collected. Most of the submodules
correspond to IANA registries."""

from . import constants, types, codes
from .constants import *
from .types import *
from .codes import *
from .optionnumbers import OptionNumber

__all__ = constants.__all__ + types.__all__ + codes.__all__ + ['OptionNumber']

media_types = {0: 'text/plain;charset=utf-8',
               16: 'application/cose;cose-type="cose-encrypt0"',
               17: 'application/cose;cose-type="cose-mac0"',
               18: 'application/cose;cose-type="cose-sign1"',
               40: 'application/link-format',
               41: 'application/xml',
               42: 'application/octet-stream',
               47: 'application/exi',
               50: 'application/json',
               51: 'application/json-patch+json',
               52: 'application/merge-patch+json',
               60: 'application/cbor',
               61: 'application/cwt',
               62: 'application/multipast-core',  # draft-ietf-core-multipart-ct
               64: 'application/link-format+cbor',  # draft-ietf-core-links-json-10
               70: 'application/oscon',  # draft-ietf-core-object-security-01
               96: 'application/cose;cose-type="cose-encrypt"',
               97: 'application/cose;cose-type="cose-mac"',
               98: 'application/cose;cose-type="cose-sign"',
               101: 'application/cose-key',
               102: 'application/cose-key-set',
               110: 'application/senml+json',
               111: 'application/sensml+json',
               112: 'application/senml+cbor',
               113: 'application/sensml+cbor',
               114: 'application/senml-exi',
               115: 'application/sensml-exi',
               256: 'application/coap-group+json',
               280: 'application/pkcs7-mime;smime-type=server-generated-key',
               281: 'application/pkcs7-mime;smime-type=certs-only',
               282: 'application/pkcs7-mime;smime-type=CMC-Request',
               283: 'application/pkcs7-mime;smime-type=CMC-Response',
               284: 'application/pkcs8',
               285: 'application/csrattrs',
               286: 'application/pkcs10',
               310: 'application/senml+xml',
               311: 'application/sensml+xml',
               1000: 'application/vnd.ocf+cbor',
               11542: 'application/vnd.oma.lwm2m+tlv',
               11543: 'application/vnd.oma.lwm2m+json',
               504: 'application/link-format+json',  # draft-ietf-core-links-json-10
               }
"""A map from CoAP-assigned integral codes to Internet media type descriptions."""

media_types_rev = {v: k for k, v in media_types.items()}
# deprecated alias. Kept alive for scripts like
# https://gitlab.f-interop.eu/f-interop-contributors/ioppytest/blob/develop/automation/coap_client_aiocoap/automated_iut.py
# that run aiocoap-client with text/plain as an argument.
#
# FIXME: Make access to this raise a DeprecationWarning
media_types_rev['text/plain'] = media_types_rev['text/plain;charset=utf-8']



import asyncio
import logging
import socket

logger = logging.getLogger(__name__)


class Endpoint:
    _scheme = None
    serializer = None

    def __init__(self, **kwargs):
        self.params = kwargs
        self._multicast = None
        self._address = None
        self._family = None
        self.is_client = False
        self.lock = asyncio.Lock()

    @property
    def multicast(self):
        return self._multicast

    @property
    def scheme(self):
        return self._scheme

    @property
    def address(self):
        return self._address[0], self._address[1]

    @address.setter
    def address(self, value):
        self._address = value

    @property
    def family(self):
        return self._family

    def is_closing(self):
        raise NotImplementedError()

    async def listen(self, server):
        raise NotImplementedError()

    def send(self, data, address, **kwargs):
        raise NotImplementedError()

    def close(self):
        raise NotImplementedError()

    def __del__(self):
        self.close()

    async def restart_transport(self, server):
        raise NotImplementedError()

    async def send_message(self, message, **kwargs):
        message.source = self.address
        logger.debug(f"Send datagram {message}")
        serializer = self.serializer()
        raw_message = serializer.serialize(message)
        self.send(bytes(raw_message), message.destination, **kwargs)



import unittest
from messages.message import Message
from messages.option import Option
from messages import Options
import defines


class Tests(unittest.TestCase):

    # def setUp(self):
    #     self.server_address = ("127.0.0.1", 5683)
    #     self.current_mid = random.randint(1, 1000)
    #     self.server_mid = random.randint(1000, 2000)
    #     self.server = CoAPServer("127.0.0.1", 5683)
    #     self.server_thread = threading.Thread(target=self.server.listen, args=(1,))
    #     self.server_thread.start()
    #     self.queue = Queue()
    #
    # def tearDown(self):
    #     self.server.close()
    #     self.server_thread.join(timeout=25)
    #     self.server = None

    def test_create_options(self):
        m = Message()
        o = Options()
        o.accept = 10000
        # setattr(o, 'accept', 10000)
        option = Option()
        option.number = defines.OptionRegistry.ACCEPT.number
        option.value = 10000



# This file is part of the Python aiocoap library project.
#
# Copyright (c) 2012-2014 Maciej Wasilak <http://sixpinetrees.blogspot.com/>,
#               2013-2014 Christian Amsüss <c.amsuess@energyharvesting.at>
#
# aiocoap is free software, this file is published under the MIT license as
# described in the accompanying LICENSE file.

"""List of known values for the CoAP "Code" field.

The values in this module correspond to the IANA registry "`CoRE Parameters`_",
subregistries "CoAP Method Codes" and "CoAP Response Codes".

The codes come with methods that can be used to get their rough meaning, see
the :class:`Code` class for details.

.. _`CoRE Parameters`: https://www.iana.org/assignments/core-parameters/core-parameters.xhtml
"""

from ...utils import ExtensibleIntEnum

# COAP_CODE_EMPTY_MESSAGE = COAP_CODE(0,00),
#   COAP_CODE_GET = COAP_CODE(0,01),
#   COAP_CODE_POST = COAP_CODE(0,02),
#   COAP_CODE_PUT = COAP_CODE(0,03),
#   COAP_CODE_DELETE = COAP_CODE(0,04),
#   COAP_CODE_201_CREATED = COAP_CODE(2,01),
#   COAP_CODE_202_DELETED = COAP_CODE(2,02),
#   COAP_CODE_203_VALID = COAP_CODE(2,03),
#   COAP_CODE_204_CHANGED = COAP_CODE(2,04),
#   COAP_CODE_205_CONTENT = COAP_CODE(2,05),
#   COAP_CODE_231_CONTINUE = COAP_CODE(2,31),
#   COAP_CODE_400_BAD_REQUEST = COAP_CODE(4,00),
#   COAP_CODE_401_UNAUTHORIZED = COAP_CODE(4,01),
#   COAP_CODE_402_BAD_OPTION = COAP_CODE(4,02),
#   COAP_CODE_403_FORBIDDEN = COAP_CODE(4,03),
#   COAP_CODE_404_NOT_FOUND = COAP_CODE(4,04),
#   COAP_CODE_405_METHOD_NOT_ALLOWED = COAP_CODE(4,05),
#   COAP_CODE_406_NOT_ACCEPTABLE = COAP_CODE(4,06),
#   COAP_CODE_408_REQUEST_ENTITY_INCOMPLETE = COAP_CODE(4,08),
#   COAP_CODE_412_PRECONDITION_FAILED = COAP_CODE(4,12),
#   COAP_CODE_413_REQUEST_ENTITY_TOO_LARGE = COAP_CODE(4,13),
#   COAP_CODE_415_UNSUPPORTED_CONTENT_FORMAT = COAP_CODE(4,15),
#   COAP_CODE_500_INTERNAL_SERVER_ERROR = COAP_CODE(5,00),
#   COAP_CODE_501_NOT_IMPLEMENTED = COAP_CODE(5,01),
#   COAP_CODE_502_BAD_GATEWAY = COAP_CODE(5,02),
#   COAP_CODE_503_SERVICE_UNAVAILABLE = COAP_CODE(5,03),
#   COAP_CODE_504_GATEWAY_TIMEOUT = COAP_CODE(5,04),
#   COAP_CODE_505_PROXYING_NOT_SUPPORTED = COAP_CODE(5,05)
class Code(ExtensibleIntEnum):
    """Value for the CoAP "Code" field.

    As the number range for the code values is separated, the rough meaning of
    a code can be determined using the :meth:`is_request`, :meth:`is_response` and
    :meth:`is_successful` methods."""

    EMPTY = 0
    GET = 1
    POST = 2
    PUT = 3
    DELETE = 4
    FETCH = 5
    PATCH = 6
    iPATCH = 7
    CREATED = 65
    DELETED = 66
    VALID = 67
    CHANGED = 68
    CONTENT = 69
    CONTINUE = 95
    BAD_REQUEST = 128
    UNAUTHORIZED = 129
    BAD_OPTION = 130
    FORBIDDEN = 131
    NOT_FOUND = 132
    METHOD_NOT_ALLOWED = 133
    NOT_ACCEPTABLE = 134
    REQUEST_ENTITY_INCOMPLETE = 136
    CONFLICT = (4 << 5) + 9
    PRECONDITION_FAILED = 140
    REQUEST_ENTITY_TOO_LARGE = 141
    UNSUPPORTED_CONTENT_FORMAT = 143
    UNSUPPORTED_MEDIA_TYPE = UNSUPPORTED_CONTENT_FORMAT  # deprecated alias
    UNPROCESSABLE_ENTITY = (4 << 5) + 22
    INTERNAL_SERVER_ERROR = 160
    NOT_IMPLEMENTED = 161
    BAD_GATEWAY = 162
    SERVICE_UNAVAILABLE = 163
    GATEWAY_TIMEOUT = 164
    PROXYING_NOT_SUPPORTED = 165

    CSM = 225
    PING = 226
    PONG = 227
    RELEASE = 228
    ABORT = 229

    def is_request(self):
        """True if the code is in the request code range"""
        return True if 1 <= self < 32 else False

    def is_response(self):
        """True if the code is in the response code range"""
        return True if 64 <= self < 192 else False

    def is_error(self):
        """True if the code is in the response code range"""
        return True if 128 <= self < 224 else False

    def is_signalling(self):
        return True if self >= 224 else False

    def is_successful(self):
        """True if the code is in the successful subrange of the response code range"""
        return True if 64 <= self < 96 else False

    def can_have_payload(self):
        """True if a message with that code can carry a payload. This is not
        checked for strictly, but used as an indicator."""
        return self.is_response() or self in (self.POST, self.PUT, self.FETCH, self.PATCH, self.iPATCH)

    @property
    def class_(self):
        """The class of a code (distinguishing whether it's successful, a
        request or a response error or more).

        >>> Code.CONTENT
        <Successful Response Code 69 "2.05 Content">
        >>> Code.CONTENT.class_
        2
        >>> Code.BAD_GATEWAY
        <Response Code 162 "5.02 Bad Gateway">
        >>> Code.BAD_GATEWAY.class_
        5
        """
        return self >> 5

    @property
    def dotted(self):
        """The numeric value three-decimal-digits (c.dd) form"""
        return "%d.%02d" % divmod(self, 32)

    @property
    def name_printable(self):
        """The name of the code in human-readable form"""
        return self.name.replace('_', ' ').title()

    def __str__(self):
        if self.is_request() or self is self.EMPTY:
            return self.name
        elif self.is_response() or self.is_signalling():
            return "%s %s" % (self.dotted, self.name_printable)
        else:
            return "%d" % self

    def __repr__(self):
        """
        >>> Code.GET
        <Request Code 1 "GET">
        >>> Code.CONTENT
        <Successful Response Code 69 "2.05 Content">
        >>> Code.BAD_GATEWAY
        <Response Code 162 "5.02 Bad Gateway">
        >>> Code(32)
        <Code 32 "32">
        """
        return '<%s%sCode %d "%s">' % ("Successful " if self.is_successful() else "",
                                       "Request " if self.is_request() else "Response " if self.is_response() else "",
                                       self, self)

    name = property(lambda self: self._name if hasattr(self, "_name") else "(unknown)",
                    lambda self, value: setattr(self, "_name", value),
                    doc="The constant name of the code (equals name_printable readable in all-caps and with underscores)")


for k in vars(Code):
    if isinstance(getattr(Code, k), Code):
        locals()[k] = getattr(Code, k)

__all__ = ['Code'] + [k for (k, v) in locals().items() if isinstance(v, Code)]



import logging
from asyncio import Protocol

from Bubot_CoAP.messages.option import Option
from . import defines
from . import utils
from .coap_protocol import CoapProtocol
from .messages.message import Message
from .serializer_tcp import SerializerTcp

logger = logging.getLogger('Bubot_CoAP')


# class RFC8323Remote:
#     """Mixin for Remotes for all the common RFC8323 processing
#
#     Implementations still need the per-transport parts, especially a
#     send_message and an _abort_with implementation.
#     """
#
#     # CSM received from the peer. The receive hook should abort suitably when
#     # receiving a non-CSM message and this is not set yet.
#     _remote_settings: Optional[Message]
#
#     # Parameter usually set statically per implementation
#     _my_max_message_size = 1024 * 1024
#
#     def __init__(self):
#         self._remote_settings = None
#
#     is_multicast = False
#     is_multicast_locally = False
#
#     # implementing interfaces.EndpointAddress
#
#     def __repr__(self):
#         return "<%s at %#x, hostinfo %s, local %s>" % (
#             type(self).__name__, id(self), self.hostinfo, self.hostinfo_local)
#
#     @property
#     def hostinfo(self):
#         # keeping _remote_hostinfo and _local_hostinfo around structurally rather than in
#         # hostinfo / hostinfo_local form looks odd now, but on the long run the
#         # remote should be able to tell the message what its default Uri-Host
#         # value is
#         return util.hostportjoin(*self._remote_hostinfo)
#
#     @property
#     def hostinfo_local(self):
#         return util.hostportjoin(*self._local_hostinfo)
#
#     @property
#     def uri_base(self):
#         if self._local_is_server:
#             raise error.AnonymousHost("Client side of %s can not be expressed as a URI" % self._ctx._scheme)
#         else:
#             return self._ctx._scheme + '://' + self.hostinfo
#
#     @property
#     def uri_base_local(self):
#         if self._local_is_server:
#             return self._ctx._scheme + '://' + self.hostinfo_local
#         else:
#             raise error.AnonymousHost("Client side of %s can not be expressed as a URI" % self._ctx._scheme)
#
#     @property
#     def maximum_block_size_exp(self):
#         if self._remote_settings is None:
#             # This is assuming that we can do BERT, so a first Block1 would be
#             # exponent 7 but still only 1k -- because by the time we send this,
#             # we typically haven't seen a CSM yet, so we'd be stuck with 6
#             # because 7959 says we can't increase the exponent...
#             #
#             # FIXME: test whether we're properly using lower block sizes if
#             # server says that szx=7 is not OK.
#             return 7
#
#         max_message_size = (self._remote_settings or {}).get('max-message-size', 1152)
#         has_blockwise = (self._remote_settings or {}).get('block-wise-transfer', False)
#         if max_message_size > 1152 and has_blockwise:
#             return 7
#         return 6  # FIXME: deal with smaller max-message-size
#
#     @property
#     def maximum_payload_size(self):
#         # see maximum_payload_size of interfaces comment
#         slack = 100
#
#         max_message_size = (self._remote_settings or {}).get('max-message-size', 1152)
#         has_blockwise = (self._remote_settings or {}).get('block-wise-transfer', False)
#         if max_message_size > 1152 and has_blockwise:
#             return ((max_message_size - 128) // 1024) * 1024 + slack
#         return 1024 + slack  # FIXME: deal with smaller max-message-size
#
#     @property
#     def blockwise_key(self):
#         return (self._remote_hostinfo, self._local_hostinfo)
#
#     # Utility methods for implementing an RFC8323 transport
#
#     def _send_initial_csm(self):
#         my_csm = Message()
#         # this is a tad awkward in construction because the options objects
#         # were designed under the assumption that the option space is constant
#         # for all message codes.
#         my_csm.type = NON
#         option = Option()
#         option.number = defines.OptionRegistry.MAX_MESSAGE_SIZE.number
#         option.value = defines.OptionRegistry.MAX_MESSAGE_SIZE.default
#         my_csm.add_option(option)
#         # block_length = optiontypes.UintOption(2, self._my_max_message_size)
#         # my_csm.opt.add_option(block_length)
#         # supports_block = optiontypes.UintOption(4, 0)
#         # my_csm.opt.add_option(supports_block)
#         self.send_message(my_csm)
#
#     def _process_signaling(self, msg):
#         if msg.code == CSM:
#             if self._remote_settings is None:
#                 self._remote_settings = {}
#             for opt in msg.opt.option_list():
#                 # FIXME: this relies on the relevant option numbers to be
#                 # opaque; message parsing should already use the appropriate
#                 # option types, or re-think the way options are parsed
#                 if opt.number == 2:
#                     self._remote_settings['max-message-size'] = int.from_bytes(opt.value, 'big')
#                 elif opt.number == 4:
#                     self._remote_settings['block-wise-transfer'] = True
#                 elif opt.number.is_critical():
#                     self.abort("Option not supported", bad_csm_option=opt.number)
#                 else:
#                     pass  # ignoring elective CSM options
#         elif msg.code in (PING, PONG, RELEASE, ABORT):
#             # not expecting data in any of them as long as Custody is not implemented
#             for opt in msg.opt.option_list():
#                 if opt.number.is_critical():
#                     self.abort("Unknown critical option")
#                 else:
#                     pass
#
#             if msg.code == PING:
#                 pong = Message(code=PONG, token=msg.token)
#                 self.send_message(pong)
#             elif msg.code == PONG:
#                 pass
#             elif msg.code == RELEASE:
#                 # The behavior SHOULD be enhanced to answer outstanding
#                 # requests, but it is unclear to which extent this side may
#                 # still use the connection.
#                 self.log.info("Received Release, closing on this end (options: %s)", msg.opt)
#                 raise CloseConnection(error.RemoteServerShutdown("Peer released connection"))
#             elif msg.code == ABORT:
#                 self.log.warning("Received Abort (options: %s)", msg.opt)
#                 raise CloseConnection(error.RemoteServerShutdown("Peer aborted connection"))
#         else:
#             self.abort("Unknown signalling code")
#
#     def abort(self, errormessage=None, bad_csm_option=None):
#         self.log.warning("Aborting connection: %s", errormessage)
#         abort_msg = Message(code=ABORT)
#         if errormessage is not None:
#             abort_msg.payload = errormessage.encode('utf8')
#         if bad_csm_option is not None:
#             bad_csm_option_option = optiontypes.UintOption(2, bad_csm_option)
#             abort_msg.opt.add_option(bad_csm_option_option)
#         self._abort_with(abort_msg)
#
#     async def release(self):
#         """Send Release message, (not implemented:) wait for connection to be
#         actually closed by the peer.
#
#         Subclasses should extend this to await closing of the connection,
#         especially if they'd get into lock-up states otherwise (was would
#         WebSockets).
#         """
#         self.log.info("Releasing connection %s", self)
#         release_msg = Message(code=RELEASE)
#         self.send_message(release_msg)
#
#         try:
#             # FIXME: we could wait for the peer to close the connection, but a)
#             # that'll need some work on the interface between this module and
#             # ws/tcp, and b) we have no peers to test this with that would
#             # produce any sensible data (as aiocoap on release just closes).
#             pass
#         except asyncio.CancelledError:
#             self.log.warning(
#                 "Connection %s was not closed by peer in time after release",
#                 self
#             )


class CoapTcpProtocol(CoapProtocol, Protocol):
    def __init__(self, server, endpoint, *, is_server=False, **kwargs):
        super(CoapTcpProtocol, self).__init__(server, endpoint)
        self._transport = None
        self.is_server = is_server
        self.remote_address = None
        self._spool = b""
        self.id = None
        if not is_server:
            self.endpoint.protocol = self

    # @property
    # def scheme(self):
    #     return self._ctx._scheme

    # def send_message(self, message: Message):
    #     # self.log.debug("Sending message: %r", message)
    #     serializer = SerializerTcp()
    #     raw_message = serializer.serialize(message)
    #     self._transport.write(bytes(raw_message))

    def send(self, data):
        logger.debug(f"Send TCP {len(data)} bytes From {self.endpoint.address[0]}: {self.endpoint.address[1]} "
                     f"To {self.remote_address[0]}:{self.remote_address[1]}")
        self._transport.write(bytes(data))

    # def _abort_with(self, abort_msg):
    #     if self._transport is not None:
    #         self.send_message(abort_msg)
    #         self._transport.close()
    #     else:
    #         # FIXME: find out how this happens; i've only seen it after nmap
    #         # runs against an aiocoap server and then shutting it down.
    #         # "poisoning" the object to make sure this can not be exploited to
    #         # bypass the server shutdown.
    #         self._ctx = None

    def connection_made(self, transport):
        self.endpoint.address = transport.get_extra_info('sockname')[:2]
        self.remote_address = transport.get_extra_info('peername')[:2]
        self.id = utils.str_append_hash(self.remote_address[0], self.remote_address[1])
        self.endpoint.pool[self.id] = self

        if self.is_server:
            logger.info(f"Server made TCP connection to {self.endpoint.address[0]}: "
                        f"{self.endpoint.address[1]} from {self.remote_address[0]}: {self.remote_address[1]}")
        else:
            logger.info(f"Client made TCP connection to {self.remote_address[0]}: "
                        f"{self.remote_address[1]} from {self.endpoint.address[0]}: {self.endpoint.address[1]}")
        self._transport = transport

        ssl_object = transport.get_extra_info('ssl_object')
        if ssl_object is not None:
            server_name = getattr(ssl_object, "indicated_server_name", None)
        else:
            server_name = None

        # `host` already contains the interface identifier, so throwing away
        # scope and interface identifier

        # def none_default_port(sockname):
        #     return (sockname[0], None if sockname[1] == self._ctx._default_port else sockname[1])
        #
        # self._local_hostinfo = none_default_port(self._local_hostinfo)
        # self._remote_hostinfo = none_default_port(self._remote_hostinfo)

        # # SNI information available
        # if server_name is not None:
        #     if self._local_is_server:
        #         self._local_hostinfo = (server_name, self._local_hostinfo[1])
        #     else:
        #         self._remote_hostinfo = (server_name, self._remote_hostinfo[1])

        if self.is_server:
            self._send_initial_csm()
        if self.server.client_manager:
            self.server.loop.create_task(self.server.client_manager.connection_made(self))

    def connection_lost(self, exc):
        if self.is_server:
            logger.info(f"Server close TCP connection to {self.endpoint.address[0]}: "
                        f"{self.endpoint.address[1]} from {self.remote_address[0]}: {self.remote_address[1]}")
        else:
            logger.info(f"Client close TCP connection to {self.remote_address[0]}: "
                        f"{self.remote_address[1]} from {self.endpoint.address[0]}: {self.endpoint.address[1]}")
        try:
            if self.server.client_manager:
                self.server.loop.create_task(self.server.client_manager.connection_lost(self, exc))
        finally:
            self.endpoint.pool.pop(self.id, None)

    def _send_initial_csm(self):
        message = Message()
        # this is a tad awkward in construction because the options objects
        # were designed under the assumption that the option space is constant
        # for all message codes.
        message.code = defines.Codes.CSM.number
        message.add_option(
            Option(defines.OptionRegistry.MAX_MESSAGE_SIZE, defines.OptionRegistry.MAX_MESSAGE_SIZE.default))
        # block_length = optiontypes.UintOption(2, self._my_max_message_size)
        # my_csm.opt.add_option(block_length)
        # supports_block = optiontypes.UintOption(4, 0)
        # my_csm.opt.add_option(supports_block)
        message.destination = self.remote_address
        message.source = self.endpoint.address
        serializer = SerializerTcp()
        raw_message = serializer.serialize(message)
        self.send(raw_message)
        # self.endpoint.send_message(message)

    def data_received(self, data):
        try:
            ...
            # A rope would be more efficient here, but the expected case is that
            # _spool is b"" and spool gets emptied soon -- most messages will just
            # fit in a single TCP package and not be nagled together.
            #
            # (If this does become a bottleneck, say self._spool = SomeRope(b"")
            # and barely change anything else).
            logger.debug(f"Recv TCP {len(data)} bytes From {self.remote_address[0]}: {self.remote_address[1]} "
                         f"To  {self.endpoint.address[0]}: {self.endpoint.address[1]} ")
            self._spool += data

            while True:
                msg_header = _extract_message_size(self._spool)
                if msg_header is None:
                    break
                msg_size = sum(msg_header)
                # if msg_size > self._my_max_message_size:
                #     self.abort("Overly large message announced") TODO
                #     return

                if msg_size > len(self._spool):
                    break

                data = self._spool[:msg_size]

                serializer = SerializerTcp()
                message = serializer.deserialize(data, self.remote_address)
                # msg.remote = self

                self._spool = self._spool[msg_size:]

                if isinstance(message, int):  # todo переделать в try catch
                    # if data[0] == b'\x16':  # client hello
                    return self.datagram_received_bad_message(message, self.remote_address)

                message.destination = self.endpoint.address
                message.multicast = False
                # message.endpoint = self.endpoint
                if self.is_server:
                    message.scheme = self.endpoint.scheme
                    message.family = self.endpoint.family
                self.message_received(message)
                return
        except RuntimeError:
            logger.exception("Exception with Executor")

    async def update_res(self, href, data, **kwargs):
        ...

    def close(self):
        if self._transport:
            self._transport.close()


def _extract_message_size(data: bytes):
    """Read out the full length of a CoAP messsage represented by data.

    Returns None if data is too short to read the (full) length.

    The number returned is the number of bytes that has to be read into data to
    start reading the next message; it consists of a constant term, the token
    length and the extended length of options-plus-payload."""

    if not data:
        return None

    l = data[0] >> 4
    tokenoffset = 2
    tkl = data[0] & 0x0f

    if l >= 13:
        if l == 13:
            extlen = 1
            offset = 13
        elif l == 14:
            extlen = 2
            offset = 269
        else:
            extlen = 4
            offset = 65805
        if len(data) < extlen + 1:
            return None
        tokenoffset = 2 + extlen
        l = int.from_bytes(data[1:1 + extlen], "big") + offset
    return tokenoffset, tkl, l


def _encode_length(l: int):
    if l < 13:
        return (l, b"")
    elif l < 269:
        return (13, (l - 13).to_bytes(1, 'big'))
    elif l < 65805:
        return (14, (l - 269).to_bytes(2, 'big'))
    else:
        return (15, (l - 65805).to_bytes(4, 'big'))



import logging
import random
import socket
import time

from .. import defines
from .. import utils
from ..messages.request import Request
from ..transaction import Transaction
from ..utils import generate_random_token

# import asyncio

__author__ = 'Giacomo Tanganelli'

logger = logging.getLogger('Bubot_CoAP')


class MessageLayer(object):
    """
    Handles matching between messages (Message ID) and request/response (Token)
    """

    def __init__(self, server, starting_mid):
        """
        Set the layer internal structure.

        :param starting_mid: the first mid used to send messages.
        """
        # self.lock = asyncio.Lock()
        self.server = server
        self._transactions = {}
        self._transactions_token = {}
        self._transactions_sent = {}
        self._transactions_sent_token = {}
        if starting_mid is not None:
            self._current_mid = starting_mid
        else:
            self._current_mid = random.randint(1, 1000)

    @staticmethod
    def fetch_token():
        return generate_random_token(8)

    def fetch_mid(self):
        """
        Gets the next valid MID.

        :return: the mid to use
        """
        current_mid = self._current_mid
        self._current_mid += 1
        self._current_mid %= 65535
        return current_mid

    def purge_sent(self, k):
        del self._transactions_sent_token[k]
        self.server.block_layer.purge_sent(k)

    def purge(self, timeout_time=defines.EXCHANGE_LIFETIME):
        for k in list(self._transactions.keys()):
            now = time.time()
            transaction = self._transactions[k]
            if transaction.timestamp + timeout_time < now or transaction.completed:
                logger.debug("Delete transaction")
                del self._transactions[k]
        for k in list(self._transactions_token.keys()):
            now = time.time()
            transaction = self._transactions_token[k]
            if transaction.timestamp + timeout_time < now or transaction.completed:
                logger.debug("Delete transaction")
                del self._transactions_token[k]
                self.server.block_layer.purge(k)

    async def receive_request(self, request):
        """
        Handle duplicates and store received messages.

        :type request: Request
        :param request: the incoming request
        :rtype : Transaction
        :return: the edited transaction
        """
        logger.info("Receive request  - " + str(request))
        try:
            host, port = request.source
        except AttributeError:
            return
        if request.multicast:
            key_mid = request.mid
            key_token = request.token  # skip duplicated from net interfaces
            if key_token in list(self._transactions_token.keys()):
                # Duplicated multicast request
                self._transactions_token[key_token].request.duplicated = True
                return self._transactions_token[key_token]
        else:
            key_mid = utils.str_append_hash(host, port, request.mid)
            key_token = utils.str_append_hash(host, port, request.token)

            if request.mid and key_mid in list(self._transactions.keys()):
                # Duplicated
                self._transactions[key_mid].request.duplicated = True
                return self._transactions[key_mid]
        request.timestamp = time.time()

        # transaction = Transaction(request=request, timestamp=request.timestamp)
        # async with transaction.lock:
        #     self._transactions[key_mid] = transaction
        #     self._transactions_token[key_token] = transaction
        ## async with self.lock:
        if key_token in self._transactions_token \
                and self._transactions_token[key_token].response is not None:  # вычитываем результат
            transaction = self._transactions_token[key_token]
            async with transaction.lock:
                self._transactions_token[key_token].request = request
                self._transactions[key_mid] = transaction
        else:
            transaction = Transaction(request=request, timestamp=request.timestamp)
            async with transaction.lock:
                self._transactions_token[key_token] = transaction
                self._transactions[key_mid] = transaction
        return transaction

    def receive_response(self, response):
        """
        Pair responses with requests.

        :type response: Response
        :param response: the received response
        :rtype : Transaction
        :return: the transaction to which the response belongs to
        """
        logger.info("Receive response - " + str(response))
        try:
            host, port = response.source
        except AttributeError:
            return
        # all_coap_nodes = defines.ALL_COAP_NODES_IPV6 if socket.getaddrinfo(host, None)[0][0] == socket.AF_INET6 else defines.ALL_COAP_NODES
        key_mid = utils.str_append_hash(host, port, response.mid)
        # key_mid_multicast = utils.str_append_hash(all_coap_nodes, port, response.mid)
        key_token = utils.str_append_hash(host, port, response.token)
        key_token_multicast = utils.str_append_hash(response.destination[0], response.destination[1], response.token)
        if response.type in [defines.Types["ACK"], defines.Types["RST"]] and key_mid in list(self._transactions_sent.keys()):
            transaction = self._transactions_sent[key_mid]
            if response.token != transaction.request.token:
                logger.warning("Tokens does not match -  response message " + str(host) + ":" + str(port))
                return None, False
        elif key_token in self._transactions_sent_token:
            transaction = self._transactions_sent_token[key_token]
        # elif key_mid_multicast in list(self._transactions_sent.keys()):
        #     transaction = self._transactions_sent[key_mid_multicast]
        elif key_token_multicast in self._transactions_sent_token:
            transaction = self._transactions_sent_token[key_token_multicast]
            if response.token != transaction.request.token:
                logger.warning("Tokens does not match -  response message " + str(host) + ":" + str(port))
                return None, False
        else:
            logger.warning(f"Un-Matched incoming response {response}")
            return None, False
        send_ack = False
        if response.type == defines.Types["CON"]:
            send_ack = True

        transaction.request.acknowledged = True
        transaction.completed = True
        transaction.response = response
        if transaction.retransmit_stop is not None:
            transaction.retransmit_stop.set()
        return transaction, send_ack

    def receive_empty(self, message):
        """
        Pair ACKs with requests.

        :type message: Message
        :param message: the received message
        :rtype : Transaction
        :return: the transaction to which the message belongs to
        """
        logger.info("Receive empty    - " + str(message))
        try:
            host, port = message.source
        except AttributeError:
            return
        all_coap_nodes = defines.ALL_COAP_NODES_IPV6 if socket.getaddrinfo(host, None)[0][
                                                            0] == socket.AF_INET6 else defines.ALL_COAP_NODES
        key_mid = utils.str_append_hash(host, port, message.mid)
        key_mid_multicast = utils.str_append_hash(all_coap_nodes, port, message.mid)
        key_token = utils.str_append_hash(host, port, message.token)
        key_token_multicast = utils.str_append_hash(all_coap_nodes, port, message.token)
        if key_mid in list(self._transactions.keys()):
            transaction = self._transactions[key_mid]
        elif key_token in self._transactions_token:
            transaction = self._transactions_token[key_token]
        elif key_mid_multicast in list(self._transactions.keys()):
            transaction = self._transactions[key_mid_multicast]
        elif key_token_multicast in self._transactions_token:
            transaction = self._transactions_token[key_token_multicast]
        else:
            logger.warning("Un-Matched incoming empty message " + str(host) + ":" + str(port))
            return None

        if message.type == defines.Types["ACK"]:
            if not transaction.request.acknowledged:
                transaction.request.acknowledged = True
            elif (transaction.response is not None) and (not transaction.response.acknowledged):
                transaction.response.acknowledged = True
        elif message.type == defines.Types["RST"]:
            if not transaction.request.acknowledged:
                transaction.request.rejected = True
            elif not transaction.response.acknowledged:
                transaction.response.rejected = True
        elif message.type == defines.Types["CON"]:
            # implicit ACK (might have been lost)
            logger.debug("Implicit ACK on received CON for waiting transaction")
            transaction.request.acknowledged = True
        else:
            logger.warning("Unhandled message type...")

        if transaction.retransmit_stop is not None:
            transaction.retransmit_stop.set()

        return transaction

    def send_request(self, request):
        """
        Create the transaction and fill it with the outgoing request.

        :type request: Request
        :param request: the request to send
        :rtype : Transaction
        :return: the created transaction
        """
        assert isinstance(request, Request)
        try:
            host, port = request.destination
        except AttributeError:
            return
        request.timestamp = time.time()
        transaction = Transaction(request=request, timestamp=request.timestamp)
        if transaction.request.type is None:
            transaction.request.type = defines.Types["CON"]

        if transaction.request.mid is None:
            transaction.request.mid = self.fetch_mid()

        if transaction.request.token is None:
            transaction.request.token = self.fetch_token()

        # logger.info("send_request - " + str(request))
        if request.multicast:
            key_token = utils.str_append_hash(request.source[0], request.source[1], request.token)
            self._transactions_sent_token[key_token] = transaction
            return self._transactions_sent_token[key_token]
        else:
            key_mid = utils.str_append_hash(host, port, request.mid)
            self._transactions_sent[key_mid] = transaction

            key_token = utils.str_append_hash(host, port, request.token)
            self._transactions_sent_token[key_token] = transaction

            return self._transactions_sent[key_mid]

    def send_response(self, transaction):
        """
        Set the type, the token and eventually the MID for the outgoing response

        :type transaction: Transaction
        :param transaction: the transaction that owns the response
        :rtype : Transaction
        :return: the edited transaction
        """
        if transaction.response is None:
            transaction.completed = True
            return
        if transaction.response.type is None:
            if transaction.request.type == defines.Types["CON"] and not transaction.request.acknowledged:
                transaction.response.type = defines.Types["ACK"]
                transaction.response.mid = transaction.request.mid
                transaction.response.acknowledged = True
                transaction.completed = True
            elif transaction.request.type == defines.Types["NON"]:
                transaction.response.type = defines.Types["NON"]
            else:
                transaction.response.type = defines.Types["CON"]
                transaction.response.token = transaction.request.token

        if transaction.response.mid is None:
            transaction.response.mid = self.fetch_mid()
            try:
                host, port = transaction.response.destination
            except AttributeError:
                return
            key_mid = utils.str_append_hash(host, port, transaction.response.mid)
            self._transactions[key_mid] = transaction

        transaction.request.acknowledged = True
        logger.info("Send response    - " + str(transaction.response))
        return transaction

    def send_empty(self, transaction, related, message):
        """
        Manage ACK or RST related to a transaction. Sets if the transaction has been acknowledged or rejected.

        :param transaction: the transaction
        :param related: if the ACK/RST message is related to the request or the response. Must be equal to
        transaction.request or to transaction.response or None
        :type message: Message
        :param message: the ACK or RST message to send
        """
        if transaction is None:
            try:
                host, port = message.destination
            except AttributeError:
                return
            key_mid = utils.str_append_hash(host, port, message.mid)
            key_token = utils.str_append_hash(host, port, message.token)
            if key_mid in self._transactions:
                transaction = self._transactions[key_mid]
                related = transaction.response
            elif key_token in self._transactions_token:
                transaction = self._transactions_token[key_token]
                related = transaction.response
            else:
                logger.info("send_empty - " + str(message))
                return message

        if message.type == defines.Types["ACK"]:
            if transaction.request == related:
                transaction.request.acknowledged = True
                transaction.completed = True
                message.mid = transaction.request.mid
                message.code = 0
                message.destination = transaction.request.source
                message.source = transaction.request.destination
                message.scheme = transaction.request.scheme
            elif transaction.response == related:
                transaction.response.acknowledged = True
                transaction.completed = True
                message.mid = transaction.response.mid
                message.code = 0
                message.token = transaction.response.token
                message.destination = transaction.response.source
                message.source = transaction.response.destination
                message.scheme = transaction.response.scheme

        elif message.type == defines.Types["RST"]:
            if transaction.request == related:
                transaction.request.rejected = True
                message._mid = transaction.request.mid
                if message.mid is None:
                    message.mid = self.fetch_mid()
                message.code = 0
                message.token = transaction.request.token
                message.destination = transaction.request.source
                message.scheme = transaction.request.scheme
            elif transaction.response == related:
                transaction.response.rejected = True
                transaction.completed = True
                message._mid = transaction.response.mid
                if message.mid is None:
                    message.mid = self.fetch_mid()
                message.code = 0
                message.token = transaction.response.token
                message.destination = transaction.response.source
                message.scheme = transaction.response.scheme
        logger.info("send_empty - " + str(message))
        return message



from .udp import UdpCoapEndpoint
# from .udp_tls2 import UdpCoapsEndpoint  #pydtls
from .udp_tls import UdpCoapsEndpoint  # aio-dtls
from .tcp import TcpCoapEndpoint
from .endpoint import Endpoint

supported_scheme = {
    'coap': UdpCoapEndpoint,
    'coaps': UdpCoapsEndpoint,
    'coap+tcp': TcpCoapEndpoint
}



# -*- coding: utf-8 -*-
import asyncio
import binascii
import random
from socket import AF_INET, AF_INET6, getaddrinfo
from urllib.parse import urlparse, SplitResult

__author__ = 'Giacomo Tanganelli'


def calc_family_by_address(address):
    if address[0] == '' or address[0] is None:
        family = AF_INET
        address = ('', address[1])
    elif address[0] == '::':
        family = AF_INET6
        address = ('[::]', address[1])
    else:
        family = getaddrinfo(address[0], address[1])[0][0]
        address = address
    return family, address


def str_append_hash(*args):
    """ Convert each argument to a lower case string, appended, then hash """
    ret_hash = ""
    for i in args:
        if isinstance(i, (str, int)):
            ret_hash += str(i).lower()
        elif isinstance(i, bytes):
            ret_hash += binascii.hexlify(i).decode("utf-8")

    return hash(ret_hash)


def check_nocachekey(option):
    """
    checks if an option is a NoCacheKey option or Etag

    :param option:
    :return:
    """
    return ((option.number & 0x1E) == 0x1C) | (option.number == 4)


def check_code(code):
    """
    checks if the response code is one of the valid ones defined in the rfc

    :param code:
    :return:
    """
    if (65 <= code <= 69) or (128 <= code <= 134) or (code == 140) or (code == 141) or (code == 143) or (
            160 <= code <= 165):
        return

    else:
        raise InvalidResponseCode


"""
exception used to signal an invalid response code
"""


class InvalidResponseCode:
    def __init__(self, code):
        self.inv_code = code


def is_uri_option(number):
    """
    checks if the option is part of uri-path, uri-host, uri-port, uri-query

    :param number:
    :return:
    """
    if number == 3 | number == 7 | number == 11 | number == 15:
        return True
    return False


def generate_random_token(size):
    return bytes([random.randint(0, 255) for _ in range(size)])


def parse_blockwise(value):
    """
    Parse Blockwise option.

    :param value: option value
    :return: num, m, size
    """

    length = byte_len(value)
    if length == 1:
        num = value & 0xF0
        num >>= 4
        m = value & 0x08
        m >>= 3
        size = value & 0x07
    elif length == 2:
        num = value & 0xFFF0
        num >>= 4
        m = value & 0x0008
        m >>= 3
        size = value & 0x0007
    else:
        num = value & 0xFFFFF0
        num >>= 4
        m = value & 0x000008
        m >>= 3
        size = value & 0x000007
    return num, int(m), pow(2, (size + 4))


def byte_len(int_type):
    """
    Get the number of byte needed to encode the int passed.

    :param int_type: the int to be converted
    :return: the number of bits needed to encode the int passed.
    """
    length = 0
    while int_type:
        int_type >>= 1
        length += 1
    if length > 0:
        if length % 8 != 0:
            length = int(length / 8) + 1
        else:
            length = int(length / 8)
    return length


def host_port_join(host, port=None):
    """Join a host and optionally port into a hostinfo-style host:port
    string

    >> host_port_join('example.com')
    'example.com'
    >> host_port_join('example.com', 1234)
    'example.com:1234'
    >> host_port_join('127.0.0.1', 1234)
    '127.0.0.1:1234'

    This is lax with respect to whether host is an IPv6 literal in brackets or
    not, and accepts either form; IP-future literals that do not contain a
    colon must be already presented in their bracketed form:

    >> host_port_join('2001:db8::1')
    '[2001:db8::1]'
    >> host_port_join('2001:db8::1', 1234)
    '[2001:db8::1]:1234'
    >> host_port_join('[2001:db8::1]', 1234)
    '[2001:db8::1]:1234'
    """
    if ':' in host and not (host.startswith('[') and host.endswith(']')):
        host = f'[{host}]'

    if port is None:
        hostinfo = host
    else:
        hostinfo = f'{host}:{port}'
    return hostinfo


def host_port_split(host_port):
    """Like urllib.parse.splitport, but return port as int, and as None if not
    given. Also, it allows giving IPv6 addresses like a netloc:

    >> host_port_split('foo')
    ('foo', None)
    >> host_port_split('foo:5683')
    ('foo', 5683)
    >> host_port_split('[::1%eth0]:56830')
    ('::1%eth0', 56830)
    """

    pseudoparsed = SplitResult(None, host_port, None, None, None)
    try:
        return pseudoparsed.hostname, pseudoparsed.port
    except ValueError:
        if '[' not in host_port and host_port.count(':') > 1:
            raise ValueError("Could not parse network location. "
                             "Beware that when IPv6 literals are expressed in URIs, they "
                             "need to be put in square brackets to distinguish them from "
                             "port numbers.")
        raise


def parse_uri2(uri):
    url = urlparse(uri)
    host, port = host_port_split(url.netloc)
    return {'scheme': url.scheme, 'address': (host, port)}


def parse_uri(uri):
    t = uri.split("://")
    tmp = t[1]
    t = tmp.split("/", 1)
    tmp = t[0]
    path = t[1]
    if tmp.startswith("["):
        t = tmp.split("]")
        host = t[0][1:]
        port = int(t[1][1:])
    else:
        t = tmp.split(":", 1)
        try:
            host = t[0]
            port = int(t[1])
        except IndexError:
            host = tmp
            port = 5683

    return str(host), port, path


def create_logging():  # pragma: no cover
    with open("logging.conf", "w") as f:
        f.writelines("[loggers]\n")
        f.writelines("keys=root\n\n")
        f.writelines("[handlers]\n")
        f.writelines("keys=consoleHandler\n\n")
        f.writelines("[formatters]\n")
        f.writelines("keys=simpleFormatter\n\n")
        f.writelines("[logger_root]\n")
        f.writelines("level=DEBUG\n")
        f.writelines("handlers=consoleHandler\n\n")
        f.writelines("[handler_consoleHandler]\n")
        f.writelines("class=StreamHandler\n")
        f.writelines("level=DEBUG\n")
        f.writelines("formatter=simpleFormatter\n")
        f.writelines("args=(sys.stdout,)\n\n")
        f.writelines("[formatter_simpleFormatter]\n")
        f.writelines("format=%(asctime)s - %(threadName)-10s - %(name)s - %(levelname)s - %(message)s\n")
        f.writelines("datefmt=")


class Tree(object):
    def __init__(self):
        self.tree = {}

    def dump(self):
        """
        Get all the paths registered in the server.

        :return: registered resources.
        """
        return sorted(list(self.tree.keys()))

    def with_prefix(self, path):
        ret = []
        for key in list(self.tree.keys()):
            if path.startswith(key):
                ret.append(key)

        if len(ret) > 0:
            return ret
        raise KeyError

    def with_prefix_resource(self, path):
        ret = []
        for key, value in self.tree.items():
            if path.startswith(key):
                ret.append(value)

        if len(ret) > 0:
            return ret
        raise KeyError

    def __getitem__(self, item):
        return self.tree[item]

    def __setitem__(self, key, value):
        self.tree[key] = value

    def __delitem__(self, key):
        del self.tree[key]

    def __contains__(self, item):
        return item in self.tree


# This file is part of the Python aiocoap library project.
#
# Copyright (c) 2012-2014 Maciej Wasilak <http://sixpinetrees.blogspot.com/>,
#               2013-2014 Christian Amsüss <c.amsuess@energyharvesting.at>
#
# aiocoap is free software, this file is published under the MIT license as
# described in the accompanying LICENSE file.

"""Tools not directly related with CoAP that are needed to provide the API"""


class ExtensibleEnumMeta(type):
    """Metaclass for ExtensibleIntEnum, see there for detailed explanations"""

    def __init__(self, name, bases, dict):
        self._value2member_map_ = {}
        for k, v in dict.items():
            if k.startswith('_'):
                continue
            if callable(v):
                continue
            if isinstance(v, property):
                continue
            instance = self(v)
            instance.name = k
            setattr(self, k, instance)
        type.__init__(self, name, bases, dict)

    def __call__(self, value):
        if isinstance(value, self):
            return value
        if value not in self._value2member_map_:
            self._value2member_map_[value] = super(ExtensibleEnumMeta, self).__call__(value)
        return self._value2member_map_[value]


class ExtensibleIntEnum(int, metaclass=ExtensibleEnumMeta):
    """Similar to Python3.4's enum.IntEnum, this type can be used for named
    numbers which are not comprehensively known, like CoAP option numbers."""

    def __add__(self, delta):
        return type(self)(int(self) + delta)

    def __repr__(self):
        return '<%s %d%s>' % (type(self).__name__, self, ' "%s"' % self.name if hasattr(self, "name") else "")

    def __str__(self):
        return self.name if hasattr(self, "name") else int.__str__(self)


def hostportjoin(host, port=None):
    """Join a host and optionally port into a hostinfo-style host:port
    string"""
    if ':' in host:
        host = '[%s]' % host

    if port is None:
        hostinfo = host
    else:
        hostinfo = "%s:%d" % (host, port)
    return hostinfo


class Timer:
    def __init__(self, interval, callback, params):
        self._ok = True
        self._task = asyncio.create_task(self._job(interval, callback, params))

    async def _job(self, interval, callback, params):
        try:
            while self._ok:
                await asyncio.sleep(interval)
                await callback(*params)
        except asyncio.CancelledError:
            return

    async def cancel(self):
        self._ok = False
        self._task.cancel()
        await self._task



# -*- coding: utf-8 -*-

# Test the support for DTLS through the SSL module. Adapted from the Python
# standard library's test_ssl.py regression test module by Björn Freise.

import sys

my_stdout = sys.stdout
import unittest
import threading
import sys
import socket
import os
import pprint

import logging

_logger = logging.getLogger(__name__)

import ssl
from dtls.wrapper import DtlsSocket
from dtls import err, force_routing_demux

# force_routing_demux()


HOST = "localhost"
CHATTY = True
CHATTY_CLIENT = True


class ThreadedEchoServer(threading.Thread):

    def __init__(self, certificate, ssl_version=None, certreqs=None, cacerts=None,
                 ciphers=None, curves=None, sigalgs=None,
                 mtu=None, server_key_exchange_curve=None, server_cert_options=None,
                 chatty=True, address=None):

        if ssl_version is None:
            ssl_version = ssl.PROTOCOL_DTLSv1_2
        if certreqs is None:
            certreqs = ssl.CERT_NONE

        self.certificate = certificate
        self.protocol = ssl_version
        self.certreqs = certreqs
        self.cacerts = cacerts
        self.ciphers = ciphers
        self.curves = curves
        self.sigalgs = sigalgs
        self.mtu = mtu
        self.server_key_exchange_curve = server_key_exchange_curve
        self.server_cert_options = server_cert_options
        self.chatty = chatty

        self.flag = None

        self.sock = DtlsSocket(socket.socket(socket.AF_INET, socket.SOCK_DGRAM),
                               keyfile=self.certificate,
                               certfile=self.certificate,
                               server_side=True,
                               cert_reqs=self.certreqs,
                               ssl_version=self.protocol,
                               ca_certs=self.cacerts,
                               ciphers=self.ciphers,
                               curves=self.curves,
                               sigalgs=self.sigalgs,
                               user_mtu=self.mtu,
                               server_key_exchange_curve=self.server_key_exchange_curve,
                               server_cert_options=self.server_cert_options
                               )

        if self.chatty:
            sys.stdout.write(' server:  wrapped server socket as %s\n' % str(address))
        self.sock.bind(address)
        self.port = self.sock.getsockname()[1]
        self.active = False
        threading.Thread.__init__(self)
        self.daemon = True

    def start(self, flag=None):
        self.flag = flag
        self.starter = threading.current_thread().ident
        threading.Thread.start(self)

    def run(self):
        self.sock.settimeout(0.05)
        self.sock.listen(0)
        self.active = True
        if self.flag:
            # signal an event
            self.flag.set()
        while self.active:
            try:
                acc_ret = self.sock.recvfrom(4096)
                if acc_ret:
                    newdata, connaddr = acc_ret
                    if self.chatty:
                        sys.stdout.write(' server:  new data from ' + str(connaddr) + '\n' + newdata.decode())
                    # self.sock.sendto(newdata.lower(), connaddr)
            except socket.timeout:
                pass
            except KeyboardInterrupt:
                self.stop()
            except Exception as e:
                if self.chatty:
                    sys.stdout.write(' server:  error ' + str(e) + '\n')
                pass
        if self.chatty:
            sys.stdout.write(' server:  closing socket as %s\n' % str(self.sock))
        self.sock.close()

    def stop(self):
        self.active = False
        if self.starter != threading.current_thread().ident:
            return
        self.join()  # don't allow spawning new handlers after we've checked


if __name__ == '__main__':
    CERTFILE = os.path.join(os.path.dirname(__file__) or os.curdir, "certs", "keycert.pem")
    CERTFILE_EC = os.path.join(os.path.dirname(__file__) or os.curdir, "certs", "keycert_ec.pem")
    ISSUER_CERTFILE = os.path.join(os.path.dirname(__file__) or os.curdir, "certs", "ca-cert.pem")
    ISSUER_CERTFILE_EC = os.path.join(os.path.dirname(__file__) or os.curdir, "certs", "ca-cert_ec.pem")
    chatty, connectionchatty = CHATTY, CHATTY_CLIENT
    server = ThreadedEchoServer(certificate=CERTFILE,
                                ssl_version=ssl.PROTOCOL_DTLSv1_2,
                                certreqs=ssl.CERT_NONE,
                                cacerts=ISSUER_CERTFILE,
                                ciphers=None,
                                curves=None,
                                sigalgs=None,
                                mtu=None,
                                server_key_exchange_curve=None,
                                server_cert_options=None,
                                chatty=chatty,
                                address=('192.168.1.13', 20102)
                                )
    flag = threading.Event()
    server.start(flag)
    server.join()



import asyncio
import logging
import random
from socket import AF_INET, AF_INET6

from aio_dtls.connection_manager.connection_manager import ConnectionManager

from bubot_helpers.ExtException import ExtException
from . import defines
from .layers.block_layer import BlockLayer
from .layers.callback_layer import CallbackLayer
from .layers.endpoint_layer import EndpointLayer
from .layers.message_layer import MessageLayer
from .layers.observe_layer import ObserveLayer
from .layers.request_layer import RequestLayer
from .layers.resource_layer import ResourceLayer
from .messages.message import Message
from .messages.request import Request
from .resources.resource import Resource
from .utils import Tree, Timer

__author__ = 'Giacomo Tanganelli'
logger = logging.getLogger('Bubot_CoAP')


class Server:
    """
    Implementation of the CoAP server
    """

    def __init__(self, *, starting_mid=None, cb_ignore_listen_exception=None, client_manager=None, **kwargs):
        """
        Initialize the server.

        :param starting_mid: used for testing purposes
        :param cb_ignore_listen_exception: Callback function to handle exception raised during the socket listen operation
        """
        self.max_retransmit = kwargs.get('max_retransmit', defines.MAX_RETRANSMIT)
        self.ask_timeout = kwargs.get('ask_timeout', defines.ACK_TIMEOUT)
        self.exchange_lifetime = kwargs.get('exchange_lifetime', defines.EXCHANGE_LIFETIME)
        self.loop = kwargs.get('loop', asyncio.get_event_loop())

        self.stopped = asyncio.Event()
        self.stopped.clear()
        self.multicast_address = {
            AF_INET: kwargs.get('multicast_ipv4', [defines.ALL_COAP_NODES]),
            AF_INET6: kwargs.get('multicast_ipv6', [defines.ALL_COAP_NODES_IPV6])
        }
        self.multicast_port = kwargs.get('multicast_port', defines.COAP_DEFAULT_PORT)
        self.to_be_stopped = []
        self.loop.create_task(self.purge())
        self.endpoint_layer = EndpointLayer(self)
        self.message_layer = MessageLayer(self, starting_mid)
        self.block_layer = BlockLayer()
        self.observe_layer = ObserveLayer()
        self.request_layer = RequestLayer(self)
        self.resource_layer = ResourceLayer(self)
        self.callback_layer = CallbackLayer(self)
        self.dtls_connection_manager = ConnectionManager(secret='test')
        # Resource directory
        root = Resource('root', self, visible=False, observable=False, allow_children=False)
        root.path = '/'
        self.client_manager = client_manager
        self.root = Tree()
        self.root["/"] = root
        self._serializer = None
        self._cb_ignore_listen_exception = cb_ignore_listen_exception

    async def purge(self):
        """
        Clean old and complete transactions

        """
        while not self.stopped.is_set():
            try:
                await asyncio.wait_for(self.stopped.wait(), 15)  # defines.EXCHANGE_LIFETIME)
            except asyncio.TimeoutError:
                pass
            self.message_layer.purge()

    async def close(self):
        """
        Stop the server.

        """
        try:
            logger.info("Stop server")
            self.stopped.set()
            for event in self.to_be_stopped:
                event.set()
            await asyncio.sleep(0.001)
            self.endpoint_layer.close()
        except Exception as err:
            raise ExtException(parent=err, action='coap server closing')

    async def receive_request(self, transaction):
        """
        Handle requests coming from the udp socket.

        :param transaction: the transaction created to manage the request
        """

        async with transaction.lock:

            transaction.separate_timer = await self._start_separate_timer(transaction)

            self.block_layer.receive_request(transaction)

            if transaction.block_transfer:
                await self._stop_separate_timer(transaction.separate_timer)
                self.message_layer.send_response(transaction)
                await self.send_datagram(transaction.response)
                return

            await self.observe_layer.receive_request(transaction)
            if transaction.resource is None:
                await self.request_layer.receive_request(transaction)

            if transaction.resource is not None and transaction.resource.changed:
                await self.notify(transaction.resource)
                transaction.resource.changed = False
            elif transaction.resource is not None and transaction.resource.deleted:
                await self.notify(transaction.resource)
                transaction.resource.deleted = False

            self.observe_layer.send_response(transaction)

            self.block_layer.send_response(transaction)

            await self._stop_separate_timer(transaction.separate_timer)

            self.message_layer.send_response(transaction)

            if transaction.response is not None:
                if transaction.response.type == defines.Types["CON"]:
                    await self.start_retransmission(transaction, transaction.response)
                if transaction.request.multicast:
                    await asyncio.sleep(random.uniform(0, 10))
                    transaction.response.source = (transaction.response.source[0], None)
                await self.send_datagram(transaction.response)
        await asyncio.sleep(0)

    async def send_message(self, message, no_response=False, endpoint=None, **kwargs):
        try:
            if isinstance(message, Request):
                if message.token is None:
                    message.token = self.message_layer.fetch_token()

                request = self.request_layer.send_request(message)
                request = self.observe_layer.send_request(request)
                request = self.block_layer.send_request(request)
                if no_response:
                    # don't add the send message to the message layer transactions
                    await self.send_datagram(request, **kwargs)
                    logger.debug(f'Send no response request {request}')
                    return
                transaction = self.message_layer.send_request(request)
                await self.send_datagram(transaction.request, endpoint=endpoint, **kwargs)
                logger.info("Send request     - " + str(request))

                if transaction.request.type == defines.Types["CON"]:
                    await self.start_retransmission(transaction, transaction.request)
                response = await self.callback_layer.wait(request, **kwargs)
                return response

            elif isinstance(message, Message):
                message = self.observe_layer.send_empty(message)
                message = self.message_layer.send_empty(None, None, message)
                await self.send_datagram(message, endpoint=endpoint, **kwargs)
        except (asyncio.TimeoutError, asyncio.CancelledError) as err:
            raise err
        except Exception as err:
            raise ExtException(parent=err)

    async def send_datagram(self, message, *, endpoint=None, **kwargs):
        """
        Send a message through the udp socket.

        :type message: Message
        :param message: the message to send
        """
        try:
            # if not self.stopped.isSet():
            # host, port = message.destination
            # host, port = message.source
            if endpoint is None:
                endpoint = self.endpoint_layer.find_sending_endpoint(message)

            if not endpoint:
                raise KeyError(
                    f'Not found endpoint for {message.source}')  # todo исключения не обрабатываются при отправке из resonse

            async with endpoint.lock:
                await endpoint.send_message(message, **kwargs)
        except Exception as err:
            raise err

    def add_resource(self, path, resource):
        """
        Helper function to add resources to the resource directory during server initialization.

        :param path: the path for the new created resource
        :type resource: Resource
        :param resource: the resource to be added
        """

        assert isinstance(resource, Resource)
        path = path.strip("/")
        paths = path.split("/")
        actual_path = ""
        i = 0
        for p in paths:
            i += 1
            actual_path += "/" + p
            try:
                res = self.root[actual_path]
            except KeyError:
                res = None
            if res is None:
                resource.path = actual_path
                self.root[actual_path] = resource
        return True

    def remove_resource(self, path):
        """
        Helper function to remove resources.

        :param path: the path for the unwanted resource
        :rtype : the removed object
        """

        path = path.strip("/")
        paths = path.split("/")
        actual_path = ""
        i = 0
        for p in paths:
            i += 1
            actual_path += "/" + p
        try:
            res = self.root[actual_path]
        except KeyError:
            res = None
        if res is not None:
            del (self.root[actual_path])
        return res

    async def start_client(self, url, **kwargs):
        return await self.endpoint_layer.start_client(url, **kwargs)

    async def add_endpoint(self, url: str, **kwargs):
        """
        Helper function to add endpoint to the endpoint directory during server initialization.

        :param endpoint: the endpoint to be added

        """
        return await self.endpoint_layer.add_by_netloc(url, **kwargs)

    #
    # def remove_endpoint(self, **kwargs):
    #     return self.endpoint_layer.remove(**kwargs)
    #     pass

    @staticmethod
    async def wait_for_retransmit_thread(transaction):
        """
        Only one retransmit thread at a time, wait for other to finish

        """
        if hasattr(transaction, 'retransmit_thread'):
            while transaction.retransmit_thread is not None:
                logger.debug("Waiting for retransmit thread to finish ...")
                await asyncio.sleep(0.01)
                continue

    async def send_block_request(self, transaction):
        """
        A former request resulted in a block wise transfer. With this method, the block wise transfer
        will be continued, including triggering of the retry mechanism.

        :param transaction: The former transaction including the request which should be continued.
        """
        transaction = self.message_layer.send_request(transaction.request)
        # ... but don't forget to reset the acknowledge flag
        transaction.request.acknowledged = False
        try:
            await self.send_datagram(transaction.request)
        except Exception as err:
            logger.error(err)
            transaction.completed = True
        if transaction.request.type == defines.Types["CON"]:
            await self.start_retransmission(transaction, transaction.request)

    async def start_retransmission(self, transaction, message):
        """
        Start the retransmission task.

        :type transaction: Transaction
        :param transaction: the transaction that owns the message that needs retransmission
        :type message: Message
        :param message: the message that needs the retransmission task
        """
        # async with transaction.lock:
        if message.type == defines.Types['CON'] and message.acknowledged and not transaction.over_tcp:
            future_time = random.uniform(defines.ACK_TIMEOUT, (defines.ACK_TIMEOUT * defines.ACK_RANDOM_FACTOR))
            transaction.retransmit_thread = self.loop.create_task(
                self._retransmit(transaction, message, future_time, 0))
            transaction.retransmit_stop = asyncio.Event()
            self.to_be_stopped.append(transaction.retransmit_stop)
            await asyncio.sleep(0.001)

    async def _retransmit(self, transaction, message, future_time, retransmit_count):
        """
        Thread function to retransmit the message in the future

        :param transaction: the transaction that owns the message that needs retransmission
        :param message: the message that needs the retransmission task
        :param future_time: the amount of time to wait before a new attempt
        :param retransmit_count: the number of retransmissions
        """
        async with transaction.lock:
            logger.debug("retransmit loop ... enter")
            while retransmit_count < defines.MAX_RETRANSMIT and (not message.acknowledged and not message.rejected) \
                    and not self.stopped.is_set():
                if transaction.retransmit_stop is not None:
                    try:
                        await asyncio.wait_for(transaction.retransmit_stop.wait(), future_time)
                    except asyncio.TimeoutError:
                        pass
                if not message.acknowledged and not message.rejected and not self.stopped.is_set():
                    retransmit_count += 1
                    future_time *= 2
                    if retransmit_count < defines.MAX_RETRANSMIT:
                        logger.debug("retransmit loop ... retransmit Request")
                        await self.send_datagram(message)

            if message.acknowledged or message.rejected:
                message.timeouted = False
            else:
                logger.warning("Give up on message {message}".format(message=message.line_print))
                message.timeouted = True
                if message.observe is not None:
                    self.observe_layer.remove_subscriber(message)

                # Inform the user, that nothing was received
                # self._callback(None)
                a = 1

            try:
                self.to_be_stopped.remove(transaction.retransmit_stop)
            except ValueError:
                pass
            transaction.retransmit_stop = None
            transaction.retransmit_thread = None
            logger.debug("retransmit loop ... exit")

    async def _start_separate_timer(self, transaction):
        """
        Start a thread to handle separate mode.

        :type transaction: Transaction
        :param transaction: the transaction that is in processing
        :rtype : the Timer object
        """
        t = Timer(defines.ACK_TIMEOUT, self.send_ack, (transaction,))
        await asyncio.sleep(0.001)
        return t

    @staticmethod
    async def _stop_separate_timer(timer):
        """
        Stop the separate Thread if an answer has been already provided to the client.

        :param timer: The Timer object
        """
        await timer.cancel()

    async def send_ack(self, transaction, message=None):
        """
        Sends an ACK message for the request.

        :param transaction: the transaction that owns the request
        """
        if message is None:
            message = transaction.request
        ack = Message()
        ack.type = defines.Types['ACK']
        async with transaction.lock:
            if not message.acknowledged and message.type == defines.Types["CON"]:
                ack = self.message_layer.send_empty(transaction, message, ack)
                if ack.type is not None and ack.mid is not None:
                    await self.send_datagram(ack)

    async def notify(self, resource):
        """
        Notifies the observers of a certain resource.

        :param resource: the resource
        """
        observers = self.observe_layer.notify(resource)
        logger.debug("Notify")
        for transaction in observers:
            with transaction:
                transaction.response = None
                transaction = self.request_layer.receive_request(transaction)
                transaction = self.observe_layer.send_response(transaction)
                transaction = self.block_layer.send_response(transaction)
                transaction = self.message_layer.send_response(transaction)
                if transaction.response is not None:
                    if transaction.response.type == defines.Types["CON"]:
                        await self.start_retransmission(transaction, transaction.response)

                    await self.send_datagram(transaction.response)



import asyncio
from unittest import IsolatedAsyncioTestCase

from Bubot_CoAP.server import Server
from Bubot_CoAP.coap_tcp_protocol import CoapProtocol


class TestTcpServer(IsolatedAsyncioTestCase):
    async def test_1(self):
        host = '192.168.1.11'
        # port = 4777
        port = 0
        server = Server()
        endpoints = await server.add_endpoint(f'coap+tcp://{host}:{port}')
        port = endpoints[0].address[1]
        loop = asyncio.get_running_loop()

        on_con_lost = loop.create_future()
        message = 'Hello World!'

        client = Server()
        transport, protocol = await loop.create_connection(
            lambda: CoapProtocol(client, None),
            host, port)

        # Wait until the protocol signals that the connection
        # is lost and close the transport.
        try:
            await on_con_lost
        finally:
            transport.close()
        ...




import logging
from ..endpoint import Endpoint, supported_scheme
from ..utils import parse_uri2


import socket

logger = logging.getLogger(__name__)


class EndpointLayer:
    """
    Class to handle the Endpoint layer
    """

    def __init__(self, server):
        self._server = server
        self.loop = self._server.loop
        self._unicast_endpoints = {}
        self._multicast_endpoints = {}

    @property
    def unicast_endpoints(self):
        return self._unicast_endpoints

    async def start_client(self, uri: str, **kwargs):
        """

        :param uri:
        :param kwargs:
        :return:
        """

        _uri = parse_uri2(uri)
        try:
            endpoint = supported_scheme[_uri['scheme']]
        except KeyError:
            raise TypeError(f'Unsupported scheme \'{_uri["scheme"]}\'')
        address = _uri['address']
        epu = endpoint(**kwargs)
        await epu.start_client(self._server, address)
        self.add(epu)

        return epu

    async def add_by_netloc(self, uri: str, **kwargs):
        """

        :param uri:
        :param kwargs:
        :return:
        """

        def extract_ip4():
            st = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            try:
                st.connect(('10.255.255.255', 1))
                ip = st.getsockname()[0]
            except Exception:
                ip = '127.0.0.1'
            finally:
                st.close()
            return ip

        def extract_ip6():
            raise NotImplementedError('default ipv6 address')
            st = socket.socket(socket.AF_INET6, socket.SOCK_DGRAM)
            try:
                st.connect(('2A00::', 1))
                ip = st.getsockname()[0]
            except Exception:
                ip = '127.0.0.1'
            finally:
                st.close()
            return ip

        async def add_all_address(_family, _result):
            try:
                addr_info = socket.getaddrinfo('', address[1], _family)
            except Exception as err:
                raise Exception(f'Bad device ip address {err}: {address[1]} {_family}')
            _new_port = None
            for addr in addr_info:
                if _new_port:
                    if _family == socket.AF_INET:
                        _address = (addr[4][0], _new_port)
                    elif _family == socket.AF_INET6:
                        _address = (addr[4][0], _new_port, addr[4][2], addr[4][3])
                    else:
                        raise NotImplementedError()
                else:
                    _address = addr[4]
                res = await endpoint.add(self, _address, **kwargs)
                if not addr[1] and not _new_port:  # if port not defined. set the same
                    _new_port = res[0].address[1]

                _result += res

        _uri = parse_uri2(uri)
        try:
            endpoint = supported_scheme[_uri['scheme']]
        except KeyError:
            raise TypeError(f'Unsupported scheme \'{_uri["scheme"]}\'')
        address = _uri['address']
        result = []

        # multicast = kwargs.get('multicast', False)
        if address[0] == '' or address[0] is None:  # IPv4 default
            address = (extract_ip4(), address[1])
            # await add_all_address(socket.AF_INET, result)
        elif address[0] == '::':
            address = (extract_ip6(), address[1])
            # await add_all_address(socket.AF_INET6, result)
        # else:

        multicast = kwargs.get('multicast', False)
        epu = endpoint(**kwargs)
        await epu.init_unicast(self._server, address)
        if self.add(epu):
            result.append(epu)
        if multicast:
            epm = endpoint(**kwargs)
            await epm.init_multicast(self._server, address)
            if self.add(epm):
                result.append(epm)
        return result

    def add(self, endpoint: Endpoint):
        """
        Handle add endpoint to server

        :type endpoint: Endpoint
        :param endpoint: the new endpoint
        :rtype : Boolean
        :return:
        """
        scheme = endpoint.scheme
        family = endpoint.family
        # await endpoint.listen(self._server)
        host, port = endpoint.address
        _endpoints = self._multicast_endpoints if endpoint.multicast else self._unicast_endpoints
        if scheme not in _endpoints:
            _endpoints[scheme] = {}
        if family not in _endpoints[scheme]:
            _endpoints[scheme][family] = {}
        if endpoint.multicast:
            if host not in _endpoints[scheme][family]:
                _endpoints[scheme][family][host] = endpoint
                return endpoint
        else:
            if host not in _endpoints[scheme][family]:
                _endpoints[scheme][family][host] = {}  # 'default': endpoint}
                if port not in self._unicast_endpoints[scheme][family][host]:
                    self._unicast_endpoints[scheme][family][host][port] = endpoint
                return endpoint
        return None

    def find_endpoint(self, *args, scheme=None, family=None, address=None):
        if scheme is None:
            scheme = self.get_first_elem(self.unicast_endpoints)
        if family is None:
            family = self.get_first_elem(self.unicast_endpoints[scheme])
        if address is None:
            address = self.get_first_elem(self.unicast_endpoints[scheme][family])
        port = self.get_first_elem(self.unicast_endpoints[scheme][family][address])
        return self.unicast_endpoints[scheme][family][address][port]

    @staticmethod
    def get_first_elem(obj):
        return list(obj.keys())[0]

    def find_sending_endpoint(self, message):
        source_address = message.source
        scheme = message.scheme
        dest_address = message.destination
        family = socket.getaddrinfo(dest_address[0], None)[0][0]
        if not source_address[0]:
            source_address = (list(self._unicast_endpoints[scheme][family].keys())[0], None)
        if source_address[0]:
            _tmp = self.unicast_endpoints[scheme][family][source_address[0]]
            if source_address[1] is None:
                return _tmp[list(_tmp.keys())[0]]
            return _tmp.get(source_address[1])
        # raise Exception('source address not defined')


    def close(self):
        def _close(endpoints):
            _keys = list(endpoints)
            for _key in _keys:
                _ep = endpoints.pop(_key)
                _ep.close()

        for scheme in list(self._unicast_endpoints):
            for family in list(self._unicast_endpoints[scheme]):
                for ip in list(self._unicast_endpoints[scheme][family]):
                    _close(self._unicast_endpoints[scheme][family].pop(ip))
                del self._unicast_endpoints[scheme][family]
            del self._unicast_endpoints[scheme]

        for scheme in list(self._multicast_endpoints):
            for family in list(self._multicast_endpoints[scheme]):
                _close(self._multicast_endpoints[scheme].pop(family))
            del self._multicast_endpoints[scheme]



import logging
import socket
import ssl

from aio_dtls import DtlsSocket

from .udp import UdpCoapEndpoint
from ..coap_udp_protocol import CoapDatagramProtocol

logger = logging.getLogger(__name__)


class UdpCoapsEndpoint(UdpCoapEndpoint):
    scheme = 'coaps'
    ssl_transport = 1  # MBEDTLS_SSL_TRANSPORT_DATAGRAM = DTLS

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        try:
            self.cert_filename = kwargs['certfile']
            self.key_filename = kwargs['keyfile']
        except KeyError as err:
            raise KeyError(f'{err} in CoapEndpoint not defined')

    def init_unicast_ip4_by_address(self, address, **kwargs):
        self._multicast = None
        self._address = address
        self._family = socket.AF_INET
        _sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        _sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self._sock = DtlsSocket(
            sock=_sock,
            endpoint=self,
            **self.params.get('socket_props', {})
        )
        self._sock.bind(address)

    def init_unicast_ip6_by_address(self, address, **kwargs):
        self._multicast = None
        self._address = address
        self._family = socket.AF_INET6
        _sock = socket.socket(socket.AF_INET6, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        _sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self._sock = DtlsSocket(
            sock=_sock,
            endpoint=self,
            **self.params.get('socket_props', {})
        )
        # self._sock = ssl.wrap_socket(
        #     _sock,
        #     # keyfile=self.key_filename, certfile=self.cert_filename,
        #     server_side=True,
        #     **self.params.get('socket_props', {})
        # )
        self._sock.bind(address)

    def send(self, data, address, **kwargs):
        self._sock.sendto(data, address, **kwargs)

    async def listen(self, server):
        await self._sock.listen(server, CoapDatagramProtocol)
        if not self.address[1]:
            self._address = self.sock.address
        logger.debug(f'run {"multicast " if self._multicast else ""}'
                     f'endpoint {self._sock.address[0]}:{self._sock.address[1]}')

    def raw_sendto(self, data, address):
        self._sock.raw_sendto(data, address)

    # async def send_alert(self, data, address, **kwargs):
    #     self._sock.send_alert(data, address, **kwargs)

    def close(self):
        if self._sock:
            self._sock.close()



__author__ = 'Mikhail Razgovorov'

import logging
import socket
import struct
from ..utils import calc_family_by_address
from .endpoint import Endpoint
from ..coap_udp_protocol import CoapDatagramProtocol
from ..defines import ALL_COAP_NODES, ALL_COAP_NODES_IPV6, COAP_DEFAULT_PORT
from ..serializer_udp import SerializerUdp

logger = logging.getLogger(__name__)


class UdpCoapEndpoint(Endpoint):
    """
    Class to handle the EndPoint.
    """
    scheme = 'coap'
    serializer = SerializerUdp

    def __init__(self, **kwargs):
        """
        Data structure that represent a EndPoint
        """
        super().__init__(**kwargs)
        self._sock = None
        self._transport = None
        self._protocol = None
        self._multicast_addresses = None

    @property
    def sock(self):
        return self._sock

    @property
    def transport(self):
        return self._transport

    # @transport.setter
    # def transport(self, value):
    #     self._transport = value

    @property
    def protocol(self):
        return self._protocol

    # @protocol.setter
    # def protocol(self, value):
    #     self._protocol = value

    @classmethod
    def init_by_socket(cls, sock: socket, is_multicast: bool = False):
        """
        Initialize the endpoint over a ready socket.

        :param sock: socket
        :param is_multicast: if socket is a multicast
        """
        self = cls()
        self._is_multicast = is_multicast
        self._sock = sock
        return self

    def send(self, data, address, **kwargs):
        self._sock.sendto(data, address)

    def _init_unicast(self, address):
        self._family, self._address = calc_family_by_address(address)
        if self._family == socket.AF_INET:  # IPv4
            self.init_unicast_ip4_by_address(address, **self.params)
        elif self._family == socket.AF_INET6:  # IPv6
            self.init_unicast_ip6_by_address(address, **self.params)

    async def init_unicast(self, server, address):
        try:
            self._init_unicast(address)
        except Exception as err:
            address = (address[0], 0)  # если порт занят сбрасываем
            self._init_unicast(address)

        await self.listen(server)
        ...

    async def init_multicast(self, server, address):
        self._family, self._address = calc_family_by_address(address)
        if self._family == socket.AF_INET:  # IPv4
            self.init_multicast_ip4_by_address(address, **self.params)
        elif self._family == socket.AF_INET6:  # IPv6
            self.init_multicast_ip6_by_address(address, **self.params)
        else:
            raise NotImplemented(f'Protocol not supported {self._family}')
        await self.listen(server)

    def init_unicast_ip4_by_address(self, address, **kwargs):
        self._multicast = None
        self._address = address
        self._family = socket.AF_INET
        self._sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        self._sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self._sock.bind(address)

    def init_unicast_ip6_by_address(self, address, **kwargs):
        self._multicast = None
        self._address = address
        self._family = socket.AF_INET6
        self._sock = socket.socket(socket.AF_INET6, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        self._sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self._sock.bind(address)

    def init_multicast_ip4_by_address(self, address, **kwargs):
        self._multicast = kwargs.get('multicast_addresses', [ALL_COAP_NODES])
        count_address = len(self._multicast)
        if not count_address:
            raise TypeError('Not defined multicast addresses')
        address = (address[0], kwargs.get('multicast_port', COAP_DEFAULT_PORT))
        self._address = address
        self._family = socket.AF_INET

        self._sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        self._sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self._sock.bind(address)
        for group in self._multicast:
            self._sock.setsockopt(
                socket.IPPROTO_IP,
                socket.IP_ADD_MEMBERSHIP,
                socket.inet_pton(socket.AF_INET, group) +
                socket.inet_pton(socket.AF_INET, address[0])
            )

    def init_multicast_ip6_by_address(self, address, **kwargs):
        self._multicast = kwargs.get('multicast_addresses', [ALL_COAP_NODES_IPV6])
        count_address = len(self._multicast)
        if not count_address:
            raise TypeError('Not defined multicast addresses')
        self._family = socket.AF_INET6
        self._address = (address[0], kwargs.get('multicast_port', COAP_DEFAULT_PORT))

        # Bugfix for Python 3.6 for Windows ... missing IPPROTO_IPV6 constant
        if not hasattr(socket, 'IPPROTO_IPV6'):
            socket.IPPROTO_IPV6 = 41

        self._sock = socket.socket(socket.AF_INET6, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        self._sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self._sock.bind(self._address)
        for group in self._multicast:
            self._sock.setsockopt(
                socket.IPPROTO_IPV6,
                socket.IPV6_JOIN_GROUP,
                socket.inet_pton(socket.AF_INET6, group) +
                struct.pack('i', address[3])
            )
        return self

    async def create_datagram_endpoint(self, server, protocol_factory):
        return await server.loop.create_datagram_endpoint(
            lambda: protocol_factory(server, self), sock=self._sock)

    async def listen(self, server):
        self._transport, self._protocol = await self.create_datagram_endpoint(server, CoapDatagramProtocol)
        _address = self._transport.get_extra_info('socket').getsockname()
        source_port = self._address[1]
        if source_port:
            if source_port != _address[1]:
                raise Exception(f'source port {source_port} not installed')
        else:
            self._address = (self._address[0], _address[1])
        logger.debug(f'run {"multicast " if self._multicast else ""}endpoint {_address[0]}:{_address[1]}')
        # _address = socket.getaddrinfo(socket.gethostname(), _address[1], socket.AF_INET, socket.SOCK_DGRAM)[0][4]

    def close(self):
        if self._transport:
            self._transport.close()
        if self._sock:
            self._sock.close()

    async def restart_transport(self, server):
        async with self.lock:
            self.close()
            await self.init_unicast(server, self.address)
            ...



# This file is part of the Python aiocoap library project.
#
# Copyright (c) 2012-2014 Maciej Wasilak <http://sixpinetrees.blogspot.com/>,
#               2013-2014 Christian Amsüss <c.amsuess@energyharvesting.at>
#
# aiocoap is free software, this file is published under the MIT license as
# described in the accompanying LICENSE file.

import abc
import collections
import struct


class OptionType(metaclass=abc.ABCMeta):
    """Interface for decoding and encoding option values

    Instances of :class:`OptionType` are collected in a list in a
    :attr:`.Message.opt` :class:`.Options` object, and provide a translation
    between the CoAP octet-stream (accessed using the
    :meth:`encode()`/:meth:`decode()` method pair) and the interpreted value
    (accessed via the :attr:`value` attribute).

    Note that OptionType objects usually don't need to be handled by library
    users; the recommended way to read and set options is via the Options
    object'sproperties (eg. ``message.opt.uri_path = ('.well-known',
    'core')``)."""

    @abc.abstractmethod
    def __init__(self, number, value):
        """Set the `self.name` and `self.value` attributes"""

    @abc.abstractmethod
    def encode(self):
        """Return the option's value in serialzied form"""

    @abc.abstractmethod
    def decode(self, rawdata):
        """Set the option's value from the bytes in rawdata"""

    @property
    def length(self):
        """Indicate the length of the encoded value"""

        return len(self.encode())


class StringOption(OptionType):
    """String CoAP option - used to represent string options. Always encoded in
    UTF8 per CoAP specification."""

    def __init__(self, number, value=""):
        self.value = value
        self.number = number

    def encode(self):
        # FIXME: actually, this should be utf8 of the net-unicode form (maybe it is)
        rawdata = self.value.encode('utf-8')
        return rawdata

    def decode(self, rawdata):
        self.value = rawdata.decode('utf-8')

    def _length(self):
        return len(self.value.encode('utf-8'))

    length = property(_length)

    def __str__(self):
        return self.value


class OpaqueOption(OptionType):
    """Opaque CoAP option - used to represent options that just have their
    uninterpreted bytes as value."""

    def __init__(self, number, value=b""):
        self.value = value
        self.number = number

    def encode(self):
        rawdata = self.value
        return rawdata

    def decode(self, rawdata):
        self.value = rawdata  # if rawdata is not None else ""

    def _length(self):
        return len(self.value)

    length = property(_length)

    def __str__(self):
        return repr(self.value)


class UintOption(OptionType):
    """Uint CoAP option - used to represent integer options."""

    def __init__(self, number, value=0):
        self.value = value
        self.number = number

    def encode(self):
        rawdata = struct.pack("!L", self.value)  # For Python >3.1 replace with int.to_bytes()
        return rawdata.lstrip(bytes([0]))

    def decode(self, rawdata):  # For Python >3.1 replace with int.from_bytes()
        value = 0
        for byte in rawdata:
            value = (value * 256) + byte
        self.value = value
        return self

    def _length(self):
        if self.value > 0:
            return (self.value.bit_length() - 1) // 8 + 1
        else:
            return 0

    length = property(_length)

    def __str__(self):
        return str(self.value)


class BlockOption(OptionType):
    """Block CoAP option - special option used only for Block1 and Block2 options.
       Currently it is the only type of CoAP options that has
       internal structure."""

    class BlockwiseTuple(collections.namedtuple('_BlockwiseTuple', ['block_number', 'more', 'size_exponent'])):
        @property
        def size(self):
            return 2 ** (self.size_exponent + 4)

        @property
        def start(self):
            return self.block_number * self.size

    def __init__(self, number, value=None):
        if value is not None:
            self._value = self.BlockwiseTuple._make(value)
        self.number = number

    value = property(lambda self: self._value,
                     lambda self, value: setattr(self, '_value', self.BlockwiseTuple._make(value)))

    def encode(self):
        as_integer = (self.value.block_number << 4) + (self.value.more * 0x08) + self.value.size_exponent
        rawdata = struct.pack("!L", as_integer)  # For Python >3.1 replace with int.to_bytes()
        return rawdata.lstrip(bytes([0]))

    def decode(self, rawdata):
        as_integer = 0
        for byte in rawdata:
            as_integer = (as_integer * 256) + byte
        self.value = self.BlockwiseTuple(block_number=(as_integer >> 4), more=bool(as_integer & 0x08),
                                         size_exponent=(as_integer & 0x07))

    def _length(self):
        return ((self.value[0].bit_length() + 3) // 8 + 1)

    length = property(_length)

    def __str__(self):
        return str(self.value)


class Enum(UintOption):
    """Enum CoAP option - used to represent integer options."""

    def __str__(self):
        return str(self.value)


import setuptools

setuptools.setup(
    name='Bubot_CoAP',
    version='2.0.0',
    license='MIT',
    author='Razgovorov Mikhail',
    author_email='1338833@gmail.com',
    url="https://github.com/businka/Bubot_CoAP.git",
    description='python library to the CoAP protocol.',
    install_requires=[
        'cbor2',
        'aio_dtls==0.0.4',
        'Bubot_Helpers>=4.0.0'
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
)



import asyncio
import logging

from bubot_helpers.ExtException import ExtException
from ..messages.request import Request
from ..messages.response import Response

logger = logging.getLogger('Bubot_CoAP')
from ..defines import MULTICAST_TIMEOUT


class CallbackLayer:
    def __init__(self, server):
        self.server = server
        self._waited_answer = {}

    async def wait(self, request: Request, *, timeout=None, **kwargs):
        # timeout = kwargs.get('timeout')
        try:
            if not timeout:
                timeout = MULTICAST_TIMEOUT
            waiter = Waiter(request, **kwargs)
            self._waited_answer[waiter.key] = waiter
            try:
                result = await asyncio.wait_for(waiter.future, timeout)
                return result
            except (asyncio.TimeoutError, asyncio.CancelledError) as err:
                if request.multicast:
                    return waiter.result
                else:
                    raise err
            except Exception as err:
                raise err
            finally:
                self._waited_answer.pop(waiter.key, None)
            pass
        except (asyncio.TimeoutError, asyncio.CancelledError) as err:
            raise err
        except Exception as err:
            raise ExtException(parent=err)

    def set_result(self, response: Response):
        try:
            waiter = self._waited_answer[response.token]
        except KeyError:
            logger.warning(f'awaited request not found {response}')
            return
        logger.debug(f'return_response - {response}')
        waiter.future = response
        pass

    def cancel_waited(self, exception):
        for key in list(self._waited_answer.keys()):
            waiter = self._waited_answer.pop(key, None)
            if not waiter.future.done():
                waiter.future.set_exception(exception)


class Waiter:
    def __init__(self, request: Request, **kwargs):
        self._request = request
        self._future = asyncio.Future()
        self._result = []

    @property
    def key(self):
        return self._request.token

    @property
    def future(self):
        return self._future

    @future.setter
    def future(self, value: Response):
        if self._request.multicast:
            self._result.append(value)
        else:
            self._future.set_result(value)

    @property
    def result(self):
        return self._result



import logging

from . import defines
from .messages.message import Message
from .messages.request import Request
from .messages.response import Response

logger = logging.getLogger('Bubot_CoAP')


class CoapProtocol:
    def __init__(self, server, endpoint):
        self.server = server
        self._transport = None
        self.endpoint = endpoint

    def message_received(self, message):
        try:
            logger.debug("Receive message - " + str(message))
            if isinstance(message, Request):
                self.server.loop.create_task(self.datagram_received_request(message))
            elif isinstance(message, Response):
                self.server.loop.create_task(self.datagram_received_response(message))
            elif message.code == defines.Codes.CSM.number:
                csm = dict(
                    max_message_size=message.get_option(defines.OptionRegistry.MAX_MESSAGE_SIZE,
                                                        defines.OptionRegistry.MAX_MESSAGE_SIZE.default)
                )
                self.endpoint.csm.set_result(csm)
                return
            else:  # is Message
                self.server.loop.create_task(self.datagram_received_message(message))

        except RuntimeError:
            logger.exception("Exception with Executor")

    def error_received(self, exc, address=None):
        logger.warning(f'protocol error received {exc}')

        # bug python https://github.com/python/cpython/issues/91227
        from sys import platform
        import asyncio
        if platform.startswith("win"):
            self.server.callback_layer.cancel_waited(asyncio.TimeoutError(str(exc)))
            self.server.loop.create_task(self.endpoint.restart_transport(self.server))
        pass

    async def datagram_received_request(self, message):
        transaction = await self.server.message_layer.receive_request(message)
        if transaction.request.duplicated and transaction.completed:
            logger.debug("message duplicated, transaction completed")
            if transaction.response is not None:
                self.server.send_datagram(transaction.response)
            return
        elif transaction.request.duplicated and not transaction.completed:
            logger.debug("message duplicated, transaction NOT completed")
            await self.server.send_ack(transaction)
            return
        await self.server.receive_request(transaction)

    async def datagram_received_response(self, message):
        transaction, send_ack = self.server.message_layer.receive_response(message)
        if transaction is None:  # pragma: no cover
            return
        await self.server.wait_for_retransmit_thread(transaction)
        if send_ack:
            await self.server.send_ack(transaction, transaction.response)
        self.server.block_layer.receive_response(transaction)
        if transaction.block_transfer:
            await self.server.send_block_request(transaction)
            return
        elif transaction is None:  # pragma: no cover
            self.server._send_rst(transaction)
            return
        self.server.observe_layer.receive_response(transaction)
        if transaction.notification:  # pragma: no cover
            ack = Message()
            ack.type = defines.Types['ACK']
            ack = self.server.message_layer.send_empty(transaction, transaction.response, ack)
            self.server.send_datagram(ack)
            self.server.callback_layer.set_result(transaction.response)
        else:
            self.server.callback_layer.set_result(transaction.response)

    async def datagram_received_message(self, message):
        transaction = self.server.message_layer.receive_empty(message)
        if transaction is not None:
            async with transaction.lock:
                self.server.block_layer.receive_empty(message, transaction)
                self.server.observe_layer.receive_empty(message, transaction)
